# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""Read and write trajectory files in various formats."""

import pkgutil

from .utils import split, formats
from .base import SuperTrajectory
from .decorators import *

# Import all trajectories
from .xyz import TrajectorySimpleXYZ, TrajectoryXYZ, TrajectoryNeighbors
from .exyz import TrajectoryEXYZ
from .pdb import TrajectoryPDB
from .hoomd import TrajectoryHOOMD
from .rumd import TrajectoryRUMD, SuperTrajectoryRUMD
from .lammps import TrajectoryLAMMPS, TrajectoryFolderLAMMPS
try:
    from .hdf5 import TrajectoryHDF5
except ImportError:
    pass
try:
    from .gsd import TrajectoryGSD
except ImportError:
    pass
from .ram import TrajectoryRam
from .dynamo import TrajectoryDynamO
from .csv import TrajectoryCSV

# Molecular trajectories
from .decorators import _Molecular
MolecularTrajectoryLAMMPS = _Molecular(TrajectoryLAMMPS)
MolecularTrajectoryXYZ = _Molecular(TrajectoryXYZ)
MolecularTrajectoryEXYZ = _Molecular(TrajectoryEXYZ)

# We build an instance of trajectory factory here and update the
# trajectory classes. Client code can update the factory with new
# classes at run time by calling update(__name__) again.
#
# The list of available formats is stored as a dict in
# Trajectory.formats, the suffixes as Trajectory.suffixes.

from .factory import TrajectoryFactory
Trajectory = TrajectoryFactory()
"""An instance of a Trajectory factory, see `TrajectoryFactory`."""
Trajectory.update(__name__)

# Update factory with plugins modules
import atooms.plugins
for _, _mod_name, _ in pkgutil.iter_modules(atooms.plugins.__path__, prefix='atooms.plugins.'):
    m = __import__(_mod_name)
    Trajectory.update(_mod_name, overwrite=False)

# Make sure the current directory is in the path. This is needed by
# the plugin mechanism. Not sure why '' should not be there btw.
import sys
__added = False
if '' not in sys.path:
    __added = True
    sys.path.append('')

# Additional plugins can be put in the atooms_plugins module
try:
    import atooms_plugins
except ImportError:
    pass
else:
    for _, _mod_name, _ in pkgutil.iter_modules(atooms_plugins.__path__,
                                                prefix='atooms_plugins.'):
        try:
            m = __import__(_mod_name)
            Trajectory.update(_mod_name, overwrite=False)
        except ImportError as err:
            # Usually it is an error in the plugin module so we show
            print(err)
            # Could not import this trajectory
            pass

# Clean up
if __added:
    sys.path.remove('')
del (__added)



import numpy


class Wall:

    """
    A wall with which particles interact.

    Examples
    --------

    .. code-block:: python

       w = Wall([1.0, 0.0, 0.0], [0,0,0])
       w.distance(numpy.array([1.0, 0.0, 0.0]))
       w.distance(numpy.array([[-2.5, 2.0, 1.0], [1.0, 0.0, 0.0]]).transpose())
    """

    def __init__(self, vector, point, species='A'):
        """The wall is a plane specified by `point` and a `vector` orthogonal to it"""
        vec = numpy.array(vector)
        self.normal = vec / numpy.sum(vec**2)**0.5
        self.point = numpy.array(point)
        self.species = species
        self._offset = numpy.dot(self.point, self.normal)

    def distance(self, point):
        """
        Return the vector distance between the `point` and the wall, where `point`
        is an array whose first dimension equals the spatial dimensionality.
        """
        scalar_distance = (numpy.dot(self.normal, point) - self._offset)
        if len(point.shape) == 1:
            return scalar_distance * self.normal
        elif len(point.shape) == 2:
            return scalar_distance * self.normal[:, None]
        else:
            raise ValueError('input vector has more than 2 dimensions')



# Boost Trajectory factory class with callbacks

import copy
from atooms.trajectory import Trajectory as __Trajectory
from atooms.trajectory.factory import TrajectoryFactory
from atooms.system import System


def _wrap_system(system):
    from atooms.trajectory.decorators import change_species
    new_system = System()
    new_system.update(system)
    new_system = change_species(new_system, 'F')
    return new_system


Trajectory = copy.deepcopy(__Trajectory)
Trajectory.register_callback(_wrap_system)

# Trajectory = TrajectoryFactory()
# for key in __Trajectory.formats:
#     new_name = __Trajectory.formats[key].__name__
#     old_cls = __Trajectory.formats[key]
#     cls = type(new_name, (old_cls, ), dict(old_cls.__dict__))
#     cls.add_class_callback(_wrap_system)
#     cls.add_self_callback(_add_interaction)
#     Trajectory.add(cls)

# # Lock the xyz format
# Trajectory.suffixes['xyz'] = Trajectory.formats['xyz']



#!/usr/bin/env python

import os
import glob
import unittest
try:
    import rumd
    from atooms.backends.rumd import RUMD
    SKIP = False
except ImportError:
    SKIP = True
from atooms.simulation import Simulation, write_thermo, write_config, target
from atooms.core.utils import setup_logging
from atooms.trajectory.rumd import TrajectoryRUMD, SuperTrajectoryRUMD

setup_logging(level=40)


class Test(unittest.TestCase):

    def setUp(self):
        if SKIP:
            self.skipTest('missing RUMD')
        self.input_file = os.path.join(os.path.dirname(__file__),
                                       '../data/ka_N256_rho1.185_rumd.xyz.gz')
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        potential.SetVerbose(False)
        potential.SetParams(i=0, j=0, Epsilon=1.0, Sigma=1.0, Rcut=2.5)
        potential.SetParams(i=1, j=0, Epsilon=1.5, Sigma=0.8, Rcut=2.5)
        potential.SetParams(i=0, j=1, Epsilon=1.5, Sigma=0.8, Rcut=2.5)
        potential.SetParams(i=1, j=1, Epsilon=0.5, Sigma=0.88, Rcut=2.5)
        self.backend = RUMD(self.input_file,
                            potentials=[potential], integrator='nvt',
                            temperature=0.80, dt=0.002)

        # Each backend must have separate potentials
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        potential.SetVerbose(False)
        potential.SetParams(i=0, j=0, Epsilon=1.0, Sigma=1.0, Rcut=2.5)
        potential.SetParams(i=1, j=0, Epsilon=1.5, Sigma=0.8, Rcut=2.5)
        potential.SetParams(i=0, j=1, Epsilon=1.5, Sigma=0.8, Rcut=2.5)
        potential.SetParams(i=1, j=1, Epsilon=0.5, Sigma=0.88, Rcut=2.5)
        self.backend_2 = RUMD(self.input_file,
                              potentials=[potential], integrator='nvt',
                              temperature=0.80, dt=0.002)

    def test_properties(self):
        t = TrajectoryRUMD(self.input_file)
        s0 = t[-1]
        sim = Simulation(self.backend,
                         output_path='/tmp/test_rumd_temp/trajectory',
                         steps=1, restart=False)
        s1 = sim.system
        self.assertAlmostEqual(s1.temperature, s0.temperature)
        self.assertAlmostEqual(s1.cell.side[0], s0.cell.side[0])
        self.assertAlmostEqual(s1.particle[0].position[0], s0.particle[0].position[0])

    def test_single(self):
        from atooms.backends.rumd import System
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_single/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        s = System(self.backend.rumd_simulation.sample)
        si.add(write_thermo, 100)
        si.add(write_config, 100)
        si.run()
        ls = glob.glob('/tmp/test_rumd_single/trajectory/*')
        self.assertEqual(len(ls), 21)

    @unittest.expectedFailure
    def test_species(self):
        """
        This test is known to fail because RUMD sample do not return the
        particle types as array and therefore they are not propagated
        when particles are not sorted by species
        """
        from atooms.backends.rumd import System
        # Create a new input file with one particle species changed
        self.input_file = os.path.join(os.path.dirname(__file__),
                                       '../data/ka_N256_rho1.185_rumd.xyz.gz')
        from atooms.core.utils import mkdir
        mkdir('/tmp/test_rumd_species/')
        with TrajectoryRUMD(self.input_file) as th:
            system = th[-1]
        system.particle[0].species = system.particle[-1].species
        with TrajectoryRUMD('/tmp/test_rumd_species/input.xyz.gz', 'w') as th:
            th.write(system)

        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_species/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        self.assertEqual(system.particle[0].species, system.particle[-1].species)
        si.add(write_config, 100)
        si.run()
        with SuperTrajectoryRUMD('/tmp/test_rumd_species/trajectory') as th:
            system = th[-1]
        self.assertEqual(system.particle[0].species, system.particle[-1].species, 'rumd requires ordered types')

    def test_multi(self):
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_multi/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        si.add(write_thermo, 50)
        si.add(write_config, 100)
        si.run()
        si.run(1000)
        ls = glob.glob('/tmp/test_rumd_multi/trajectory/*')
        self.assertEqual(len(ls), 31)
        self.assertEqual(si.current_step, 3000)
        tmp = open('/tmp/test_rumd_multi/trajectory.thermo', 'r').readlines()
        steps = int(tmp[-1].split()[0])
        self.assertEqual(steps, 3000)
        self.assertEqual(len(tmp), 61 + 1)  # one is for comment line

    def test_multi_writing(self):
        # Test that we cumulate current_step and configurations
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_multi_writing/trajectory',
                        enable_speedometer=False)
        si.add(write_config, 500)
        si.run(10000)
        si.run(5000)
        ls = glob.glob('/tmp/test_rumd_multi_writing/trajectory/*')
        self.assertEqual(len(ls), 31)
        self.assertEqual(si.current_step, 15000)

        # This second simulation tests that the folder gets cleared up
        si = Simulation(self.backend, steps=5000,
                        output_path='/tmp/test_rumd_multi_writing/trajectory',
                        enable_speedometer=False)
        si.add(write_config, 500)
        si.run()
        si.run()
        ls = glob.glob('/tmp/test_rumd_multi_writing/trajectory/*')
        self.assertEqual(len(ls), 21)
        self.assertEqual(si.current_step, 10000)

    def test_multi_2(self):
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_multi_2/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        si.add(write_thermo, 100)
        si.add(write_config, 100)
        si.run()
        si = Simulation(self.backend_2,
                        output_path='/tmp/test_rumd_multi_2/trajectory',
                        steps=1000, checkpoint_interval=100,
                        restart=False)
        si.add(write_thermo, 100)
        si.add(write_config, 100)
        si.run()

    def test_multi_rmsd(self):
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_multi_rmsd/trajectory',
                        checkpoint_interval=100, steps=1000000000,
                        restart=False)
        si.add(write_thermo, 100)
        si.add(write_config, 100)
        si.add(target, 1000, 'rmsd', 1.0)
        si.run()

    def test_system_copy(self):
        """Make sure rumd systems can be deepcopied"""
        import copy
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_single/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        s0 = copy.copy(si.system)
        s1 = copy.deepcopy(si.system)
        self.assertEqual(si.system.particle[0].position[-1], s1.particle[0].position[-1])

    def test_potential(self):
        import copy
        from atooms.backends.rumd import RUMD
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_single/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        pos0 = si.system.particle[0].position[0]
        s = copy.deepcopy(si.system)
        self.assertAlmostEqual(si.system.potential_energy(per_particle=True), -6.4014371344819665, 3)
        si.run(100)
        pos1 = si.system.particle[0].position[0]
        si.system = s
        si.run(100)
        pos2 = si.system.particle[0].position[0]
        self.assertTrue(abs(pos1 - pos0) > 1e-2)
        self.assertTrue(abs(pos1 - pos2) < 1e-4)

    def test_copy_and_run(self):
        self.skipTest('skipped test')
        import copy
        import atooms.trajectory.ram
        from atooms.backends.rumd import RUMD
        import rumd
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        potential.SetVerbose(False)
        potential.SetParams(i=0, j=0, Epsilon=1.0, Sigma=1.0, Rcut=2.5)
        potential.SetParams(i=1, j=0, Epsilon=1.5, Sigma=0.8, Rcut=2.5)
        potential.SetParams(i=0, j=1, Epsilon=1.5, Sigma=0.8, Rcut=2.5)
        potential.SetParams(i=1, j=1, Epsilon=0.5, Sigma=0.88, Rcut=2.5)
        input_file = os.path.join(os.path.dirname(__file__),
                                  '../data/ka_N256_rho1.185_rumd.xyz.gz')
        backend = RUMD(input_file, integrator='nvt', temperature=0.8, dt=0.002, forcefield=[potential])
        sim = Simulation(backend,
                         output_path='/tmp/test_rumd_single/trajectory',
                         steps=100, restart=False)

        # First set if runs
        # print sim.system.particle[0].position[0]
        tr = atooms.trajectory.ram.TrajectoryRamFull()
        th = atooms.trajectory.ram.TrajectoryRamFull()
        tr[0] = sim.system
        # print tr[0].particle[0].position[0]
        sim.run(1000)  # sometimes is nan, depending on steps

        # Ram does not work with rumd because of change in System init
        th[0] = sim.system
        # print sim.system.particle[0].position, sim.system.potential_energy(per_particle=True)

        sim.run(100)
        # print sim.system.particle[0].position[0]
        # this assignment leads to trouble, BUT ONLY IF WE STORE THE SAMPLES IN tr TRAJECTORY
        # assigning tr[0] above also leads to a different trajectory...!!
        sim.system = th[0]
        # sim.system = tr[0]
        # print sim.system.particle[0].position, sim.system.potential_energy(per_particle=True)

    def test_unfold(self):
        from atooms.backends.rumd import unfold

        def unf(sim):
            unfold(sim.system).particle[0].position
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_single/trajectory',
                        steps=2000, restart=False)
        si.add(unf, 100)
        si.run()

    def test_leakage(self):
        # self.skipTest('skipped test')
        from atooms.backends.rumd import System
        from atooms.trajectory.ram import TrajectoryRamFull
        from atooms.backends.rumd import unfold
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_single/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        # self.backend.rumd_simulation.sample.__swig_destroy__(self.backend.rumd_simulation.sample)
        # del self.backend.rumd_simulation.sample

        # This does not leak memory
        trj = TrajectoryRamFull()
        trj[0] = si.system
        for i in range(5):
            si.system = trj[0]
            si.run()
            unfold(si.system).particle[0].position[0], unfold(trj[0]).particle[0].position[0]

        # This does not leak memory anymore because we use System.update()
        # trj = TrajectoryRamFull()
        # for i in range(5):
        #     trj[0] = si.system
        #     si.run()
        #     #print si.system.particle[0].position[0], trj[0].particle[0].position[0]
        #     print unfold(si.system).particle[0].position[0], unfold(trj[0]).particle[0].position[0]

    def test_update(self):
        # self.skipTest('skipped test')
        from atooms.backends.rumd import System
        from atooms.trajectory.ram import TrajectoryRamFull
        si = Simulation(self.backend,
                        output_path='/tmp/test_rumd_single/trajectory',
                        steps=2000, checkpoint_interval=100,
                        restart=False)
        itg = si.system.sample.GetIntegrator()
        trj = TrajectoryRamFull()
        trj[0] = si.system
        si.system.thermostat.temperature = 10.0
        # Objects are different but underlying information is shared...
        # print trj[0].sample.GetIntegrator().GetInfoString(18), itg.GetInfoString(18)
        self.assertNotEqual(trj[0].sample.GetIntegrator(), itg)
        # self.assertEqual(trj[0].sample.GetIntegrator().GetInfoString(18), itg.GetInfoString(18))
        trj[0] = si.system  # This uses update
        # print trj[0].sample.GetIntegrator().GetInfoString(18), itg.GetInfoString(18)
        self.assertNotEqual(trj[0].sample.GetIntegrator(), itg)

    def test_new_format(self):
        input_file = os.path.join(os.path.dirname(__file__),
                                  '../data/ka_rumd_N1200.xyz.gz')
        with TrajectoryRUMD(input_file) as th:
            self.assertAlmostEqual(th[0].cell.side[0], 10.03)
            self.assertAlmostEqual(th[0].cell.side[1], 10.03)
            self.assertAlmostEqual(th[0].cell.side[2], 10.03)

    def tearDown(self):
        os.system('rm -rf /tmp/test_rumd_*')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""Test trajectory decorators."""

import os
import unittest
import numpy

from atooms.trajectory import TrajectoryXYZ


def _equal(system1, system2):
    for p1, p2 in zip(system1.particle, system2.particle):
        if p1.species != p2.species:
            print(p1.species, p2.species)
            return False
    return True


class TestDecorators(unittest.TestCase):

    Trajectory = TrajectoryXYZ

    def setUp(self):
        self.finp = '/tmp/test_decorators.xyz'
        with open(self.finp, 'w') as fh:
            fh.write("""\
2
step:1 cell:6.0,6.0,6.0
A 1.0 -1.0 0.0
A 2.9 -2.9 0.0
2
step:2 cell:6.0,6.0,6.0
A 1.1 -1.1 0.0
A -2.9 -2.9 0.0
2
step:3 cell:6.0,6.0,6.0
A 1.2 -1.2 0.0
A -2.9 2.9 0.0
2
step:4 cell:6.0,6.0,6.0
A 1.2 -1.2 0.0
A -2.9 2.9 0.0
""")

    def test_unfolded_jump(self):
        from atooms.trajectory.decorators import Unfolded
        with self.Trajectory(self.finp) as th:
            with Unfolded(th) as tu:
                self.assertEqual(list(tu[2].particle[1].position), [3.1, -3.1, 0.0])  # unfolded
                self.assertEqual(list(th[2].particle[1].position), [-2.9, 2.9, 0.0])  # original

    def test_sliced(self):
        from atooms.trajectory.decorators import Sliced
        with Sliced(self.Trajectory(self.finp), slice(0, 2, 1)) as th:
            self.assertEqual(th.steps, [1, 2])
            self.assertEqual(list(th[-1].particle[0].position), [1.1, -1.1, 0.0])
            self.assertEqual(list(th[-1].particle[1].position), [-2.9, -2.9, 0.0])

    def test_callback_args(self):
        def cbk_1(system, scale):
            for p in system.particle:
                p.position *= scale
            return system

        def cbk_2(system, offset):
            for p in system.particle:
                p.position -= offset
            return system

        def cbk_3(system, memory):
            """Callback that modifies a mutable."""
            memory.append(1)
            return system

        memo = []
        with self.Trajectory(self.finp) as th:
            th.register_callback(cbk_1, 2.0)
            th.register_callback(cbk_2, offset=1.0)
            th.register_callback(cbk_3, memo)
            self.assertEqual(th.steps, [1, 2, 3, 4])
            self.assertEqual(list(th[-1].particle[0].position),
                             [1.2 * 2 - 1.0, -1.2 * 2 - 1.0, 0.0 * 2 - 1.0])
            self.assertEqual(memo, [1])

    def test_change_species(self):
        from copy import deepcopy
        from atooms.system import System, Particle
        system_A = System([Particle(species='A'), Particle(species='B')])
        system_C = System([Particle(species='0'), Particle(species='1')])
        system_F = System([Particle(species='1'), Particle(species='2')])
        from atooms.trajectory.decorators import change_species
        # DO nothing here
        self.assertTrue(_equal(system_A, change_species(deepcopy(system_A), 'A')))
        self.assertTrue(_equal(system_C, change_species(deepcopy(system_C), 'C')))
        self.assertTrue(_equal(system_F, change_species(deepcopy(system_F), 'F')))
        # Change
        self.assertTrue(_equal(system_C, change_species(deepcopy(system_A), 'C')))
        self.assertTrue(_equal(system_F, change_species(deepcopy(system_A), 'F')))
        self.assertTrue(_equal(system_A, change_species(deepcopy(system_C), 'A')))
        self.assertTrue(_equal(system_F, change_species(deepcopy(system_C), 'F')))
        self.assertTrue(_equal(system_A, change_species(deepcopy(system_F), 'A')))
        self.assertTrue(_equal(system_C, change_species(deepcopy(system_F), 'C')))

    def tearDown(self):
        os.remove(self.finp)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import sys
import random
import copy
import numpy
import unittest
from atooms.system import *


class Test(unittest.TestCase):

    def setUp(self):
        N = 100
        L = 10.0
        random.seed(1)
        self.ref = System()
        self.ref.cell = Cell([L, L, L])
        self.ref.particle = []
        self.ref.thermostat = Thermostat(1.0)
        self.ref.barostat = Barostat(1.0)
        self.ref.reservoir = Reservoir(1.0)
        while len(self.ref.particle) <= N:
            pos = [(random.random() - 0.5) * L,
                   (random.random() - 0.5) * L,
                   (random.random() - 0.5) * L]
            self.ref.particle.append(Particle(position=pos))

    def test_dict(self):
        pos = self.ref['position'].copy()
        self.ref['position'] = pos*2
        self.assertEqual(self.ref['position'][0, 0], pos[0, 0]*2)
        self.assertEqual(self.ref['position'][0, 0], self.ref.particle[0].position[0])

    def test_ndim(self):
        system = System()
        self.assertEqual(system.number_of_dimensions, 0)
        system.cell = Cell([2.0, 2.0])
        self.assertEqual(system.number_of_dimensions, 2)
        particle = Particle(position=[0, 0])
        self.assertEqual(particle.velocity.shape[0], 2)

    def test_density(self):
        system = copy.copy(self.ref)
        density_old = system.density
        system.density = density_old * 1.1
        self.assertAlmostEqual(system.density, density_old * 1.1)

        # When there are no particles, density is zero
        system = System()
        self.assertAlmostEqual(system.density, 0.0)

    def test_density_cluster(self):
        system = copy.deepcopy(self.ref)
        system.cell = None
        self.assertAlmostEqual(system.density, 0.1, 1)

    def test_temperature(self):
        system = copy.copy(self.ref)
        system.set_temperature(1.0)
        self.assertAlmostEqual(system.temperature, 1.0)
        system.scale_velocities(1.0)
        self.assertAlmostEqual(system.temperature, 1.0)

        # Pathological case
        system.particle = system.particle[0: 1]
        self.assertAlmostEqual(system.temperature, 2.0 / system.number_of_dimensions * system.kinetic_energy())

        # Empty system
        system = System()
        self.assertAlmostEqual(system.temperature, 0.0)

    def _test_cm(self, order='F'):
        system = copy.copy(self.ref)
        system.set_temperature(1.0)
        system.view('position', order=order)
        system.view('velocity', order=order)
        for i in range(system.number_of_dimensions):
            self.assertAlmostEqual(system.center_of_mass('velocity')[i], system.cm_velocity[i])
            self.assertAlmostEqual(system.center_of_mass('position')[i], system.cm_position[i])
        system.fix_momentum()
        for i in range(system.number_of_dimensions):
            self.assertAlmostEqual(system.center_of_mass('velocity')[i], 0.0)
        pos_cm = system.cm_position
        for p in system.particle:
            p.position -= pos_cm
        for i in range(system.number_of_dimensions):
            self.assertAlmostEqual(system.center_of_mass('position')[i], 0.0)

    def test_cm(self):
        self._test_cm('C')
        self._test_cm('F')

    def test_pbc_center(self):
        system = copy.copy(self.ref)
        # Move the center of the cell so that positions are within 0 and L
        system.cell.center = system.cell.side / 2
        for p in system.particle:
            p.position += system.cell.side / 2
        # Check that distances are the same
        for i in range(len(system.particle)):
            self.assertAlmostEqual(sum(self.ref.particle[0].distance(self.ref.particle[i], self.ref.cell)**2),
                                   sum(system.particle[0].distance(system.particle[i], system.cell)**2))
        # Move one particle out of the box and fold it back
        pos = copy.copy(system.particle[0].position)
        system.particle[0].position += system.cell.side
        system.particle[0].fold(system.cell)
        self.assertAlmostEqual(pos[0], system.particle[0].position[0])
        self.assertAlmostEqual(pos[1], system.particle[0].position[1])
        self.assertAlmostEqual(pos[2], system.particle[0].position[2])

    def test_fold(self):
        system = copy.copy(self.ref)
        pos = copy.copy(system.particle[0].position)
        system.particle[0].position += system.cell.side
        system.particle[0].fold(system.cell)
        self.assertAlmostEqual(pos[0], system.particle[0].position[0])
        self.assertAlmostEqual(pos[1], system.particle[0].position[1])
        self.assertAlmostEqual(pos[2], system.particle[0].position[2])

    def test_overlaps(self):
        from atooms.system.particle import overlaps
        system = copy.copy(self.ref)
        for p in system.particle:
            p.radius = 1e-10
        pos = copy.copy(system.particle[1].position)
        system.particle[0].position = pos
        ov, ipart = overlaps(system.particle, system.cell)
        self.assertTrue(ov)
        self.assertEqual(ipart, [(0, 1)])

    def test_dump(self):
        self.assertEqual(self.ref.dump('spe')[-1],
                         self.ref.dump('particle.species')[-1])
        self.assertAlmostEqual(self.ref.dump('pos')[-1][-1],
                               self.ref.dump('particle.position')[-1][-1])
        self.assertAlmostEqual(self.ref.dump('vel')[-1][-1],
                               self.ref.dump('particle.velocity')[-1][-1])

    def test_species(self):
        system = copy.copy(self.ref)
        npart = len(system.particle)
        for p in system.particle[0: 10]:
            p.species = 'B'
        for p in system.particle[10: 30]:
            p.species = 'C'
        from atooms.system.particle import composition, distinct_species
        self.assertEqual(distinct_species(system.particle), ['A', 'B', 'C'])
        self.assertEqual(system.distinct_species, ['A', 'B', 'C'])
        self.assertEqual(composition(system.particle)['A'], npart - 30)
        self.assertEqual(composition(system.particle)['B'], 10)
        self.assertEqual(composition(system.particle)['C'], 20)

        # Unhashable numpy scalars
        for p in system.particle:
            p.species = numpy.array(1)
        self.assertEqual(system.distinct_species, [numpy.array(1)])
        for p in system.particle:
            p.species = numpy.array('A')
        self.assertEqual(system.distinct_species, [numpy.array('A')])

    def test_concentration(self):
        system = System(N=100)
        system.set_concentration({'A': 0.2, 'B': 0.8})
        self.assertEqual(system.composition, {'A': 20, 'B': 80})
        system.set_concentration({'A': 1/3, 'B': 2/3})
        self.assertEqual(system.composition, {'A': 33, 'B': 67})
        system = System(N=101)
        system.set_concentration({'A': 0.2, 'B': 0.8})
        self.assertEqual(system.composition, {'A': 20, 'B': 81})
        system = System(N=250)
        system.set_concentration({'A': 0.2, 'B': 0.7, 'C': 0.1})
        self.assertEqual(system.composition, {'A': 50, 'B': 175, 'C': 25})

        def _():
            system.set_concentration({'A': 0.1, 'B': 1.0})
        self.assertRaises(AssertionError, _)

    def test_species_layout(self):
        system = copy.copy(self.ref)
        for p in system.particle[0: 10]:
            p.species = 'B'
        for p in system.particle[10: 30]:
            p.species = 'C'
        self.assertTrue(system.species_layout == 'A')
        system.species_layout = 'C'
        self.assertTrue(system.species_layout == 'C')
        system.species_layout = 'F'
        self.assertTrue(system.species_layout == 'F')
        system.species_layout = 'A'
        self.assertTrue(system.species_layout == 'A')

    def test_packing(self):
        import math
        system = copy.copy(self.ref)
        self.assertAlmostEqual(system.packing_fraction * 6 / math.pi, system.density)

    def test_gyration(self):
        from atooms.system.particle import gyration_radius
        system = copy.copy(self.ref)

        # Ignore cell
        rg1 = gyration_radius(system.particle, method='N1')
        rg2 = gyration_radius(system.particle, method='N2')
        self.assertAlmostEqual(rg1, rg2)

        # With PBC all estimates are different but bounds must be ok
        rg1 = gyration_radius(system.particle, system.cell, method='min')
        rg2 = gyration_radius(system.particle, system.cell, method='N1')
        rg3 = gyration_radius(system.particle, system.cell, method='N2')
        self.assertLessEqual(rg1, rg2)
        self.assertLessEqual(rg3, rg2)

        # Equilateral triangle
        system.particle = [Particle(), Particle(), Particle()]
        system.particle[0].position = numpy.array([0.0, 0.0, 0.0])
        system.particle[1].position = numpy.array([1.0, 0.0, 0.0])
        system.particle[2].position = numpy.array([0.5, 0.5*3**0.5, 0])
        # Put the triangle across the cell
        system.particle[0].position -= 1.01*system.cell.side/2
        system.particle[1].position -= 1.01*system.cell.side/2
        system.particle[2].position -= 1.01*system.cell.side/2
        system.particle[0].fold(system.cell)
        system.particle[1].fold(system.cell)
        system.particle[2].fold(system.cell)
        rg1 = gyration_radius(system.particle, system.cell, method='min')
        rg2 = gyration_radius(system.particle, system.cell, method='N1')
        rg3 = gyration_radius(system.particle, system.cell, method='N2')
        self.assertAlmostEqual(rg1, 0.57735026919)
        self.assertAlmostEqual(rg2, 0.57735026919)
        self.assertAlmostEqual(rg3, 0.57735026919)

    def test_interaction(self):
        from atooms.system.interaction import InteractionBase
        system = copy.copy(self.ref)
        self.assertAlmostEqual(system.potential_energy(), 0.0)
        system.interaction = InteractionBase()
        system.interaction.compute('energy', system.dump('position'))
        self.assertAlmostEqual(system.potential_energy(), 0.0)
        self.assertAlmostEqual(system.potential_energy(normed=True), 0.0)
        self.assertAlmostEqual(system.total_energy(), system.kinetic_energy())

    def test_interaction_add(self):
        from atooms.system.interaction import InteractionBase
        x, y = InteractionBase(), InteractionBase()
        x.energy, y.energy = 1., 1.
        z = sum([x, y])
        self.assertAlmostEqual(z.energy, 2.)
        # Setting energy to None in one interaction obliterates
        # the calculation of the sum of the energies
        y.energy = None
        z = x + y
        self.assertEqual(z.energy, None)

    def test_interaction_terms(self):
        from atooms.system.interaction import InteractionBase, Interaction
        x, y = InteractionBase(), InteractionBase()
        z = Interaction(x, y)
        w = Interaction()
        w.add(x)
        w.add(y)
        w.compute('forces', position=numpy.array([[1.]]))
        self.assertAlmostEqual(w.energy, 0.0)
        self.assertTrue(w.forces is not None)

    def test_interaction_terms_via_system(self):
        from atooms.system.interaction import InteractionBase, Interaction
        system = copy.copy(self.ref)
        system.interaction = Interaction(InteractionBase())
        system.compute_interaction('forces')
        self.assertAlmostEqual(system.potential_energy(), 0.0)

    def test_interaction_wall(self):
        from atooms.simulation import Simulation, Scheduler
        from atooms.backends.f90 import Interaction as InteractionParticle

        model = {
            "potential": [
                {
                    "type": "lennard_jones",
                    "parameters": {"epsilon": [[1.0]], "sigma": [[1.0]]}
                }
            ],
            "cutoff": [
                {
                    "type": "cut_shift",
                    "parameters": {"rcut": [[2.5]]}
                }
            ]
        }

        N = 9
        L = 10.0
        system = System()
        # We still define a big supercell, but we should not need it
        system.cell = Cell([L * 10, L * 10, L * 10])
        system.particle = []
        for i in range(N):
            system.particle.append(Particle(position=[1 + i, 1 + i, 0.0],
                                            velocity=numpy.random.normal(0.0, 0.1, 3),
                                            species=1))
        system.wall = [
            Wall([+1, 0, 0], [0, 0, 0]),
            Wall([-1, 0, 0], [L, 0, 0]),
            Wall([0, +1, 0], [0, 0, 0]),
            Wall([0, -1, 0], [0, L, 0]),
        ]

        # Particle-wall potential
        def inverse_power(rsq, u, w, h, sigma=1.0, epsilon=1.0, exponent=12):
            u[:] = epsilon * (sigma**2 / rsq[:])**(exponent // 2)
            w[:] = exponent * u[:] / rsq[:]

        # We have two interaction terms
        system.interaction = Interaction()
        system.interaction.add(InteractionWall(system.wall, inverse_power))
        system.interaction.add(InteractionParticle(model))
        self.assertAlmostEqual(system.potential_energy(per_particle=True), 0.07016883058835639)

    def test_interaction_field(self):
        from atooms.system.interaction import InteractionField, Interaction

        system = copy.copy(self.ref)

        def gravitational_field(pos, mass, g=9.81):
            u = mass[:] * g * pos[-1, :]
            grad = numpy.zeros_like(pos)
            grad[-1, :] = mass[:] * g
            return u, grad, None
        system.interaction = InteractionField(gravitational_field,
                                              variables={'pos': 'particle.position',
                                                         'mass': 'particle.mass'})
        system.compute_interaction('forces')
        # print(system.potential_energy())
        # print(system.interaction.forces[:, 0])
        self.assertAlmostEqual(system.interaction.forces[0, 0], 0)
        self.assertAlmostEqual(system.interaction.forces[1, 0], 0)
        self.assertAlmostEqual(system.interaction.forces[2, 0], -9.81)

    def test_overlap(self):
        from atooms.system.particle import self_overlap, collective_overlap
        sys1 = copy.deepcopy(self.ref)
        sys2 = copy.deepcopy(self.ref)
        sys1.particle = sys1.particle[:int(len(sys1.particle) / 2)]
        sys2.particle = sys2.particle[int(len(sys2.particle) / 2):]
        self.assertEqual(0, self_overlap(sys1.particle, sys2.particle, 0.001))
        self.assertEqual(0, collective_overlap(sys1.particle, sys2.particle, 0.001, sys1.cell.side))
        sys1.particle = sys1.particle
        sys2.particle = sys1.particle
        self.assertEqual(1, self_overlap(sys1.particle, sys2.particle, 0.001))
        self.assertEqual(1, collective_overlap(sys1.particle, sys2.particle, 0.001, sys1.cell.side))

    def test_overlap_random(self):
        # This test may fail from time to time
        from atooms.system.particle import collective_overlap
        N = 1000
        L = 5.0
        sys = [System(), System()]
        sys[0].cell = Cell([L, L, L])
        sys[1].cell = Cell([L, L, L])
        sys[0].particle = []
        sys[1].particle = []
        for _ in range(N):
            pos = [(random.random() - 0.5) * L,
                   (random.random() - 0.5) * L,
                   (random.random() - 0.5) * L]
            sys[0].particle.append(Particle(position=pos))
        for _ in range(N):
            pos = [(random.random() - 0.5) * L,
                   (random.random() - 0.5) * L,
                   (random.random() - 0.5) * L]
            sys[1].particle.append(Particle(position=pos))
        a = 0.3
        q_rand = ((a**3 * 4./3*3.1415) * N / sys[0].cell.volume)
        self.assertTrue(abs(q_rand - collective_overlap(sys[0].particle, sys[1].particle, a, sys[0].cell.side)) < 0.5)

    def test_view(self):
        import numpy
        from atooms.system import Particle, System

        p = [Particle(), Particle()]
        s = System(p)
        pos = s.dump("pos", order='F', view=True)

        # Modify the dumped array in place preserves the view
        pos[:, 0] += 1.0
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        # Modify the position array in place preserves the view
        p[0].position *= 2
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        # Modify the position array in place preserves the view
        p[0].position[:] = p[0].position[:] + 4
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        pos[:, 0] = pos[:, 0] + 1.0
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        # Reassining the position will of course destroy the view
        p[0].position = p[0].position * 2
        self.assertFalse((p[0].position == pos[:, 0]).all())
        self.assertFalse(numpy.may_share_memory(p[0].position, pos[:, 0]))

    def test_view_new(self):
        self.skipTest('')
        import numpy
        from atooms.system import Particle, System

        p = [Particle(), Particle()]
        s = System(p)
        pos = s._view("pos", order='F')

        # Modify the dumped array in place preserves the view
        pos[:, 0] += 1.0
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        # Modify the position array in place preserves the view
        p[0].position *= 2
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        # Modify the position array in place preserves the view
        p[0].position[:] = p[0].position[:] + 4
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        pos[:, 0] = pos[:, 0] + 1.0
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))
        # Reassining the position will of course destroy the view
        p[0].position = p[0].position * 2
        self.assertFalse((p[0].position == pos[:, 0]).all())
        self.assertFalse(numpy.may_share_memory(p[0].position, pos[:, 0]))

    def test_view_clear(self):
        import numpy
        from atooms.system import Particle, System

        p = [Particle(), Particle()]
        s = System(p)

        # We check that particle positions are views on dump array
        pos = s.dump("pos", order='F', view=True)
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))

        # We should get the same dump array
        pos = s.dump("pos", order='F', view=True)
        self.assertTrue((p[0].position == pos[:, 0]).all())
        self.assertTrue(numpy.may_share_memory(p[0].position, pos[:, 0]))

        # We clear the dump array
        pos1 = s.dump("pos", order='F', view=True, clear=True)
        pos1 += 1
        self.assertFalse((p[0].position == pos[:, 0]).all())
        self.assertFalse(numpy.may_share_memory(p[0].position, pos[:, 0]))

    def test_dump_cbk(self):
        """
        Make sure that applying callbacks or changing the particle array
        size creates a new dump.
        """
        import numpy
        from atooms.system import Particle, System

        p = [Particle(), Particle()]
        s = System(p)
        pos1 = s.view("pos", order='C')
        pos2 = s.view("pos", order='F')
        self.assertEqual(pos1.shape, (2, 3))
        self.assertEqual(pos2.shape, (3, 2))

        for view in [False, True]:
            p = [Particle(), Particle()]
            s = System(p)
            pos1 = s.dump("pos", view=view)

            def cbk(system):
                s = copy.copy(system)
                s.particle = [system.particle[0]]
                return s
            s = cbk(s)
            pos2 = s.dump('pos', view=view)
            self.assertEqual(pos1.shape, (2, 3))
            self.assertEqual(pos2.shape, (1, 3))
            # Grandcanonical
            s.particle.append(Particle())
            self.assertEqual(s.dump('pos', view=view).shape, (2, 3))
            # Reassign particle
            # Expected failure with view = True
            if not view:
                s.particle[0] = Particle(position=[1.0, 1.0, 1.0])
                self.assertEqual(s.dump('pos', view=view)[0][0], 1.0)

    def test_dump_cbk_new(self):
        """
        Make sure that applying callbacks or changing the particle array
        size creates a new dump.
        """
        self.skipTest('')
        import numpy
        from atooms.system import Particle, System

        for view in [False, True]:
            p = [Particle(), Particle()]
            s = System(p)
            if view:
                pos1 = s._view("pos")
            else:
                pos1 = s._dump("pos")

            def cbk(system):
                s = copy.copy(system)
                s.particle = [system.particle[0]]
                return s
            s = cbk(s)
            if view:
                pos2 = s._view('pos')
            else:
                pos2 = s._dump('pos')
            self.assertEqual(pos1.shape, (2, 3))
            self.assertEqual(pos2.shape, (1, 3))
            # Grandcanonical
            s.particle.append(Particle())
            if view:
                self.assertEqual(s._view('pos').shape, (2, 3))
            else:
                self.assertEqual(s._dump('pos').shape, (2, 3))
            # Reassign particle
            # Expected failure with view = True
            # But it should work with dump... not if we have dump = view.copy()
            if not view:
                s.particle[0] = Particle(position=[1.0, 1.0, 1.0])
                self.assertEqual(s._dump('pos')[0][0], 1.0)

    def test_slab(self):
        from atooms.system import Particle, System

        s = System(N=256)
        s.density = 1.0
        slabs = s.slab(axis=2, width=1.0)
        slabs = s.slab(axis=2, n=5)
        # slabs[-1].show(now=True)
                
    def test_dump_species(self):
        """
        Make sure that changing species in the dump is reflected in the
        particle species and viceversa.
        """
        import numpy
        from atooms.system import Particle, System

        view = True
        p = [Particle(), Particle()]
        s = System(p)
        spe = s.dump("particle.species", view=True)
        spe[0] = 'B'
        self.assertEqual(spe[0], s.particle[0].species)
        # With this syntax, the numpy scalar preserves the view!
        # We should upgrade species to property and hide this inside
        s.particle[0].species[()] = 'C'
        self.assertEqual(spe[0], s.particle[0].species)

    def test_dumps(self):
        """Check that dump order does not matter"""
        import numpy
        from atooms.system import Particle, System

        p = [Particle(), Particle()]
        s = System(p)
        # View, dump, view: we should preserve the view and get a copy of the dump
        pos1 = s.dump("pos", view=True)
        pos1[0, 0] = 0.0
        pos2 = s.dump('pos', view=False)
        pos2[0, 0] = 1.0
        pos3 = s.dump("pos", view=True)
        self.assertNotAlmostEqual(pos1[0, 0], pos2[0, 0])
        self.assertAlmostEqual(pos1[0, 0], pos3[0, 0])
        self.assertFalse(numpy.may_share_memory(pos1, pos2))
        self.assertTrue(numpy.may_share_memory(pos1, pos3))

    def test_dump_flatten(self):
        """Check that flattening a dump does not change successive views"""
        import numpy
        from atooms.system import Particle, System

        p = [Particle(), Particle()]
        s = System(p)
        self.assertEqual(len(s.dump('pos', view=True).shape), 2)
        self.assertEqual(len(s.dump('pos', view=False, flat=True).shape), 1)
        self.assertEqual(len(s.dump('pos', view=True).shape), 2)
        self.assertEqual(len(s.dump('pos', view=False, flat=True).shape), 1)

    def test_dump_fail(self):
        """Check that dump fails on unknown attributes"""
        for what in ['whatever', 'particle.whatever', 'cell.whatever']:
            try:
                self.ref.dump(what)
            except AttributeError:
                pass

    def test_decimate(self):
        from atooms.system import Particle, System
        from atooms.system.particle import composition, decimate
        p = [Particle(species='A')]*20 + [Particle(species='B')]*10
        pnew = decimate(p, 12)
        x = composition(pnew)
        self.assertEqual(x['A'], 8)
        self.assertEqual(x['B'], 4)

    @unittest.skipIf(sys.version_info.major == 2, 'skip show() tests with python 2')
    def test_show(self):
        self.skipTest('')
        N = 3
        L = 5.0
        system = System()
        system.cell = Cell([L, L, L])
        system.particle = []
        for _ in range(N):
            pos = (numpy.random.random(len(system.cell.side)) - 0.5) * system.cell.side
            p = Particle(position=pos)
            system.particle.append(p)

        try:
            system.show(backend='')
        except ValueError:
            pass
        try:
            system.show(backend='matplotlib')
            system.show(backend='ovito')
            system.show(backend='3dmol')
        except ImportError:
            self.skipTest('missing backend')

    def test_rotate(self):
        """Rotate particles so that the principal axis is along the y axis"""
        from atooms.system import Particle, Cell
        from atooms.system.particle import rotate
        p1, p2, p3 = Particle(position=[0.0, 0.0, 0.0]), Particle(position=[1.0, 0.0, 0.0]), Particle(position=[2.0, 0.0, 0.0])
        particle = [p1, p2, p3]
        cell = Cell(side=[10.0, 10.0, 10.0])
        rotated = rotate(particle, cell)
        self.assertAlmostEqual(rotated[0].position[1], 0, 6)
        self.assertAlmostEqual(rotated[1].position[1], 1, 6)
        self.assertAlmostEqual(rotated[2].position[1], 2, 6)

    def test_init(self):
        """Test __init__() syntax"""
        self.assertEqual(len(System(N=1).particle), 1)
        self.assertEqual(len(System(N=10).particle), 10)
        self.assertEqual(len(System(N=100).particle), 100)
        self.assertEqual(System(N=10).composition, {'A': 10})
        self.assertEqual(System(N={'B': 4}).composition, {'B': 4})
        self.assertEqual(System(N={'Si': 4, 'O': 2}).composition, {'Si': 4, 'O': 2})
        self.assertEqual(System(N={'A': 4, 'B': 10, 'C': 2}).composition, {'A': 4, 'B': 10, 'C': 2})
        self.assertAlmostEqual(System(N={'Si': 10, 'O': 20}).concentration['Si'], 10/30.)
        self.assertAlmostEqual(System(N={'Si': 10, 'O': 20}).concentration['O'], 20/30.)

    def test_replicate(self):
        s = System(N=2**3, d=2)
        L = s.cell.side[0]
        for p in s.particle:
            p.radius = 0.1
        s.replicate(3, 0)
        x = s.dump('pos', order='F')
        self.assertAlmostEqual(min(x[0, :]), -max(x[0, :]))
        self.assertAlmostEqual(s.cell.side[0], L * 3)

        s = System(N=2**3, d=3)
        L = s.cell.side[0]
        s.replicate(4, 1)
        x = s.dump('pos', order='F')
        self.assertAlmostEqual(min(x[1, :]), -max(x[1, :]))
        self.assertAlmostEqual(s.cell.side[1], L * 4)
        # s.show(outfile='1.png')

    def test_molecule(self):
        from atooms.system.molecule import Molecule
        particle = [Particle(position=[0.00000, -0.06556, 0.00000], species=1),  # , charge=-0.834),
                    Particle(position=[0.75695, 0.52032, 0.00000], species=2),  # , charge=-0.417),
                    Particle(position=[-0.75695, 0.52032, 0.00000], species=2)]  # , charge=-0.417)]
        molecule = Molecule(particle, bond=[(0, 0), (0, 1)], angle=[(1, 0, 2)])
        self.assertTrue(all(numpy.isclose(molecule.center_of_mass, [0., 0.32502667, 0.])))
        molecule.center_of_mass = [0., 0., 0.]
        self.assertTrue(all(numpy.isclose(molecule.center_of_mass, [0., 0., 0.])))

        from atooms.system.particle import _lattice
        lattice = _lattice(4)
        cell = Cell(numpy.ones(3))

        molecules = []
        for p in lattice:
            m = Molecule(particle, bond=[(0, 0), (0, 1)], angle=[(1, 0, 2)])
            m.center_of_mass = p.position
            molecules.append(m)

        system = System(molecule=molecules)
        self.assertEqual(len(system.particle), 12)
        self.assertEqual(len(system.molecule), 4)
        self.assertTrue(all(numpy.isclose(system.dump('molecule.center_of_mass')[0],
                                          [0.25, -0.25, -0.25])))
        self.assertEqual(system.dump('molecule.species')[0], '122')
        self.assertTrue(numpy.all(numpy.isclose(system.dump('particle.position')[0],
                                                [0.25, -0.64058667, -0.25])))
        self.assertTrue(numpy.all(numpy.isclose(system.dump('molecule.bond')[0],
                                                [[0, 0], [0, 1]])))
        self.assertTrue(numpy.all(numpy.isclose(system.dump('molecule.orientation')[0],
                                                system.dump('molecule.orientation')[1])))
        self.assertTrue(all(numpy.isclose(system.dump('molecule.orientation')[0][0],
                                          [0, -0.39058667, 0])))
        # print(system.dump('molecule.orientation')[0])
        # This was in commit 27893294 and failed
        # [-0.75695, 0.58588, 0])))

        # Check custom orientation vectors
        self.assertTrue(numpy.all([system.molecule[0].orientation_vector('CM-1'),
                                   system.molecule[0].orientation_vector('CM-2'),
                                   system.molecule[0].orientation_vector('CM-3')] ==
                                  system.molecule[0].orientation))

        # Just a formal test that vector product is parsed, I have not checked
        # it is correct (it gives [0., 0., 0.88696373])
        system.molecule[0].orientation_vector('1-2x2-3')
        o = system.molecule[0].orientation_vector('1-2x2-3', normed=True)
        self.assertAlmostEqual(numpy.linalg.norm(o), 1.0)

        self.assertTrue(numpy.all(system.molecule[0].custom_orientation(['1-2']) ==
                                  numpy.array([system.molecule[0].orientation_vector('1-2')])))

        # This test assumes there are no pbc involved
        self.assertTrue(numpy.all(system.molecule[0].orientation_vector('e2e') ==
                                  system.molecule[0].particle[len(particle)-1].position -
                                  system.molecule[0].particle[0].position))

        # print(system.dump('molecule.orientation')[0])

        # This is how it could look like
        try:
            from atooms.backends.f90 import InteractionMolecular
            interaction = InteractionMolecular(model)
            bnds = system.view('molecule.bond')
            angs = system.view('molecule.angle')
            interaction.compute('forces', box, pos, ids, bnds, angs)  # , dihs, imps)
        except:
            pass

    def test_molecule_2d(self):
        from atooms.system.molecule import Molecule
        particle = [Particle(position=[0.00000, -0.06556], species=1),
                    Particle(position=[0.75695, 0.52032], species=2),
                    Particle(position=[-0.75695, 0.52032], species=2)]
        molecule = Molecule(particle, bond=[(0, 0), (0, 1)], angle=[(1, 0, 2)])
        self.assertTrue(all(numpy.isclose(molecule.center_of_mass, [0., 0.32502667])))
        molecule.center_of_mass = [0., 0.]
        self.assertTrue(all(numpy.isclose(molecule.center_of_mass, [0., 0.])))

        # Test without cell
        from atooms.system.particle import _lattice
        lattice = _lattice(4, d=2, spacing=3.0)
        cell = Cell([5.0, 5.0])
        molecules = []
        for p in lattice:
            m = Molecule(particle, bond=[(0, 0), (0, 1)], angle=[(1, 0, 2)])
            m.center_of_mass = p.position
            molecules.append(m)

        system = System(molecule=molecules)
        for p in system.particle:
            p.radius = 0.1
        system.cell = cell
        # system.show(now=True)

        # Test with cell
        from atooms.system.particle import _lattice
        lattice = _lattice(4, d=2, spacing=3.0)
        cell = Cell([5.0, 5.0])
        m = Molecule(particle, bond=[(0, 0), (0, 1)], angle=[(1, 0, 2)], cell=cell)
        m.center_of_mass = numpy.array([2.0, 0.0])
        system = System(molecule=[m], cell=cell)
        for p in system.particle:
            p.radius = 0.1

        # Show molecule and its orientations
        try:
            import matplotlib.pyplot as plt
            fig = system.show()
            ax = fig.axes[0]
            ax.plot(*m.center_of_mass, 'o', color='black')
            for o in m.orientation:
                ax.arrow(*m.center_of_mass, *o, color='black')
            # plt.show()
        except ImportError:
            pass

    def test_molecule_2d_pbc(self):
        from atooms.system.molecule import Molecule
        from atooms.system.particle import _lattice, Particle

        L = 5.0
        particle = [Particle(position=[x, 0.0]) for x in [-1.0, 0.0, 1.0, L/2-0.1]]
        particle = [Particle(position=[x, 0.0]) for x in [-L/2+1, -L/2+0.1, L/2-0.5]]
        # m.center_of_mass = numpy.array([0.0, 0.0])
        lattice = _lattice(4, d=2, spacing=3.0)        
        cell = Cell([L, L])
        m = Molecule(particle, cell=cell, bond=[(i, i+1) for i in range(len(particle))])
        # m.center_of_mass = numpy.array([0.0, 0.0])
        system = System(molecule=[m], cell=cell)
        for p in system.particle:
            p.radius = 0.1

        e2e = m.orientation_vector('e2e')
        try:
            import matplotlib.pyplot as plt
            fig = system.show()
            ax = fig.axes[0]
            ax.plot(*m.center_of_mass, 'o', color='black')
            ax.arrow(*m.particle[0].position, *e2e, color='black')
            #for o in m.orientation:
            #    ax.arrow(*m.center_of_mass, *o, color='black')
            plt.show()
        except ImportError:
            pass
            
if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""Point particles in a cartesian reference frame."""

import logging
import numpy
import random
from copy import deepcopy
from atooms.core import ndim as _ndim

_log = logging.getLogger(__name__)


class Particle(object):

    def __init__(self, position=None, velocity=None, species='A',
                 mass=1.0, radius=0.5):
        self.species = species
        """The chemical species of the particle."""
        self.mass = mass
        self.radius = radius
        if position is None:
            self.position = numpy.zeros(_ndim)
        else:
            self.position = numpy.asarray(position)
        if velocity is None:
            self.velocity = numpy.zeros_like(self.position)
        else:
            self.velocity = numpy.asarray(velocity)

    @property
    def diameter(self):
        """Particle diameter."""
        return self.radius * 2

    def nearest_image(self, particle, cell, copy=False, folded=False):
        """
        Return the nearest image of `particle` in the given `cell`.

        If `copy` is `False`, the particle is transformed into to its
        nearest image, otherwise the fucction returns a copy of the
        nearest image particle and leave the original particle as is.

        If `folded` is True, the coordinates are assumed to be folded
        into the central cell (or in a cell just next to it),
        otherwise they will be assumed to lie in an arbitrary periodic
        cell.
        """
        rij = self.position - particle.position
        if folded:
            rij = _periodic_vector(rij, cell.side)
        else:
            rij = _periodic_vector_unfolded(rij, cell.side)
        if copy:
            image = deepcopy(self)
            image.position = particle.position + rij
            return image
        else:
            self.position[:] = particle.position + rij
            return self

    def nearest_image_of(self, particle, cell, folded=False):
        """
        Return a copy of the nearest image of `particle` from `self` in
        the given `cell`.

        If `folded` is True, the coordinates are assumed to be folded
        into the central cell (or in a cell just next to it),
        otherwise they will be assumed to lie in an arbitrary periodic
        cell.
        """
        rij = particle.position - self.position
        if folded:
            rij = _periodic_vector(rij, cell.side)
        else:
            rij = _periodic_vector_unfolded(rij, cell.side)
        image = deepcopy(particle)
        image.position = self.position + rij
        return image

    def distance(self, particle, cell=None, folded=True):
        """
        Return the distance from another `particle`.

        If `cell` is given, return the distance from the nearest image
        of `particle`. In this case, if `folded` is True, the
        coordinates are assumed to be folded into the central cell (or
        in a cell just next to it), otherwise they will be assumed to
        lie in an arbitrary periodic cell.

        If `cell` is None, periodic boundary conditions are *not*
        applied and folded is irrelevant.
        """
        r = self.position - particle.position
        if cell is not None:
            # Apply periodic boundary conditions
            if folded:
                r = _periodic_vector(r, cell.side)
            else:
                r = _periodic_vector_unfolded(r, cell.side)
        return r

    def __repr__(self):
        return 'Particle(species={0.species}, mass={0.mass}, ' \
            'position={0.position}, velocity={0.velocity}, ' \
            'radius={0.radius})'.format(self)

    def fold(self, cell):
        """Fold self into central cell."""

        # Move the center to 0
        self.position -= cell.center
        self.position[:] = _periodic_vector_unfolded(self.position, cell.side)

        # Restore the center
        self.position += cell.center
        return self

    def maxwellian(self, T):
        """
        Assign the velocity to particle according to a Maxwell-Boltzmann
        distribution at temperature `T`.
        """
        for i in range(len(self.velocity)):
            self.velocity[i] = random.gauss(0, numpy.sqrt(T / self.mass))

    @property
    def kinetic_energy(self):
        """Kinetic energy."""
        return 0.5 * self.mass * numpy.dot(self.velocity, self.velocity)


# Utility functions

def _periodic_vector(vec, box):
    # TODO: what about particle distances precisely equal to L/2 or -L/2?
    for i in range(vec.shape[0]):
        if vec[i] > box[i] / 2:
            vec[i] += - box[i]
        elif vec[i] < -box[i] / 2:
            vec[i] += box[i]
    # Compact version
    # return numpy.where(abs(a) > box/2, a-numpy.copysign(box, a), a)
    return vec


def _periodic_vector_unfolded(vec, box):
    return vec - numpy.rint(vec / box) * box
    # Optimized version
    # return vec - numpy.rint(vec * invbox) * box


def _periodic_vector_delta_unfolded(vec, box):
    """
    Return periodic vector delta to enable in-place modification of
    folded particles.
    """
    # Optimized version
    # return numpy.rint(vec * invbox) * box
    return numpy.rint(vec / box) * box


def fix_total_momentum(particles):
    """
    Subtract out the center of mass velocity from a list of
    `particles`.
    """
    vcm = cm_velocity(particles)
    for p in particles:
        p.velocity -= vcm
    return particles


def cm_velocity(particle):
    """Velocity of the center of mass of a list of particles."""
    if len(particle) == 0:
        return 0.0
    vcm = numpy.zeros_like(particle[0].velocity)
    mtot = 0.0
    for p in particle:
        vcm += p.velocity * p.mass
        mtot += p.mass
    return vcm / mtot


def cm_position(particle):
    """Center-of-mass of a list of particles."""
    rcm = numpy.zeros_like(particle[0].position)
    mtot = 0.0
    for p in particle:
        rcm += p.position * p.mass
        mtot += p.mass
    return rcm / mtot


def distinct_species(particles):
    """Return sorted list of distinct `species` of `particles`."""
    try:
        return list(sorted({p.species for p in particles}))
    except TypeError:
        # This is (most likely) an unhashable numpy array
        # We accept both int and str species labels
        dtype = particles[0].species.dtype
        # We cast species as strings
        species = list(sorted({str(p.species) for p in particles}))
        # Then restore their original types
        return [numpy.array(_, dtype=dtype) for _ in species]


def composition(particles):
    """
    Return a dictionary containing the number of particles of each
    species appearing the input `particles` list.
    """
    from collections import defaultdict
    comp = defaultdict(int)
    for p in particles:
        try:
            comp[p.species] += 1
        except TypeError:
            comp[int(p.species)] += 1
    return comp


def rotate(particle, cell):
    """
    Return a list of particles rotated around the main symmetry axis
    of the set.
    """
    import math

    def norm2(vec):
        return numpy.dot(vec, vec)

    def rotation(axis, theta):
        """
        Return the rotation matrix associated with counterclockwise rotation about
        the given axis by theta radians.
        """
        axis = numpy.asarray(axis)
        theta = numpy.asarray(theta)
        axis = axis / math.sqrt(numpy.dot(axis, axis))
        a = math.cos(theta / 2.0)
        b, c, d = -axis * math.sin(theta / 2.0)
        aa, bb, cc, dd = a * a, b * b, c * c, d * d
        bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
        return numpy.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

    p = deepcopy(particle)
    dist = [sum(p[0].distance(pi, cell)**2) for pi in p]
    dmax = max(dist)
    imax = dist.index(dmax)
    z_axis = numpy.array([0, -1, 0])
    pr_axis = p[0].position - p[imax].position
    ro_axis = numpy.cross(pr_axis, z_axis)
    theta = math.acos(numpy.dot(pr_axis, z_axis) / (norm2(pr_axis) * norm2(z_axis))**0.5)
    for pi in p:
        pi.position = numpy.dot(rotation(ro_axis, theta), pi.position)
    return p


def overlaps(particle, cell):
    """Check presence of overlaps between particles."""
    x = []
    for i, pi in enumerate(particle):
        for j, pj in enumerate(particle):
            if j <= i:
                continue
            d = sum(pi.distance(pj, cell)**2)**0.5
            if d < (pi.radius + pj.radius):
                x.append((i, j))
    return len(x) > 0, x


def gyration_radius(particles, cell=None, weight=None, center=None,
                    method="N1"):
    """
    Gyration radius of a list of `particles` in a `cell` with an
    optional `weight`.

    The optional `center` variable is the index of the central
    particle for pbc unfolding. If `weight` is given but `center` is
    not, the latter will be the index of the first largest element of
    `weight`.

    Possible values of `methods` are `N1` (default), `N2` and `min`.
    """
    if method not in ['N1', 'N2', 'min']:
        raise ValueError('unknown method %s' % method)

    # Order N^2 calculation
    if method == 'N2':
        rg = 0.0
        for pi in particles:
            for pj in particles:
                if pi is pj:
                    continue
                dr = pi.distance(pj, cell)
                rg += numpy.dot(dr, dr)
        return (rg / (2*len(particles)**2))**0.5

    # Order N^1 calculation
    elif method == 'N1':
        # Assign weights
        if weight is None:
            weight = numpy.ones(len(particles))
            if center is None:
                center = 0
        else:
            if len(weight) != len(particles):
                raise ValueError('n. weights differs from n. particles')
            weight = numpy.asarray(weight)
            if center is None:
                center = weight.argmax()

        # Unfold cluster across pbc
        p_central = particles[center]
        if cell is None:
            cluster = particles
        else:
            cluster = []
            for p in particles:
                cluster.append(p.nearest_image(p_central, cell, copy=True))

        def weighted_cm_position(particle, weight):
            """Weighted center-of-mass of a list of particles."""
            rcm = numpy.zeros_like(particle[0].position)
            wtot = sum(weight)
            for i, p in enumerate(particle):
                rcm += p.position * weight[i]
            return rcm / wtot

        # Compute gyration radius
        rcm = weighted_cm_position(cluster, weight)
        rg = 0.0
        for i, p in enumerate(cluster):
            dr = p.position - rcm
            rg += numpy.dot(dr, dr) * weight[i]
        rg /= len(cluster)
        return rg**0.5

    # Minimize over possible centers
    elif method == 'min':
        rg = None
        for center in range(len(particles)):
            rg_new = gyration_radius(particles, cell=cell,
                                     weight=weight, center=center,
                                     method="N1")
            if rg is None:
                rg = rg_new
            else:
                rg = min(rg_new, rg)
        return rg


def collective_overlap(particle, other, a, side, normalize=True):
    """Compute collective overlap between two lists of particles."""
    # These optimizations via numpy are necessary. Computing O(N^2)
    # quantities in pure python takes forever.
    x = numpy.array([p.position for p in particle])
    y = numpy.array([p.position for p in other])
    dr = numpy.ndarray(x.shape[0])
    rij = numpy.asarray(y)
    q = 0
    for i in range(y.shape[0]):
        dr = x[:, :] - y[i, :]
        dr = dr - numpy.rint(dr / side) * side
        dr = numpy.sum(dr**2, axis=1)
        q += (dr < a**2).sum()
    if normalize:
        q /= float(len(particle))
    return q


def self_overlap(particle, other, a, normalize=True):
    """Compute self overlap between two lists of unfolded particles."""
    # This quantity is O(N) and therefore we use pure python
    rij = [sum(p.distance(o)**2) for p, o in zip(particle, other)]
    q = (numpy.array(rij) < a**2).sum()
    if normalize:
        q /= float(len(particle))
    return q


def decimate(particle, N):
    """
    Return a decimated list of N particles keep the same chemical
    composition as in the input list `particle`.
    """
    import random
    if N > len(particle):
        raise ValueError('cannot increase number of particles')
    x = composition(particle)
    scale = float(N) / len(particle)
    idx = {}
    for species in x:
        idx[species] = random.sample(range(x[species]), int(x[species] * scale))

    # if sum([len(idx[species]) for species in x]) != N:
    #     print sum([len(idx[species]) for species in x]), N,  [len(idx[species]) for species in x], x
    #     raise ValueError('cannot preserve composition')

    # Go through the particles once and map sequential indices to the subset of indices of a given species
    from collections import defaultdict
    idx_species = defaultdict(list)
    for i, p in enumerate(particle):
        idx_species[p.species].append(i)

    # Collect selected particles
    pnew = []
    for species in sorted(x):
        for i in idx[species]:
            ii = idx_species[species][i]
            pnew.append(particle[ii])

    return pnew


def _lattice(N, d=3, spacing=-1.0):
    """
    Return a list of `N` particles on the sites of a simple cubic
    lattice in `d`-dimensional space. A perfect crystalline
    configuration is etched so as to match the input value of `N`. The
    particles' position are between -0.5 and 0.5 along each axis.
    """
    import random

    # Find minimum number of particles to match input N
    n = 1
    while n**d < N:
        n += 1

    # Put particles on a cubic lattice
    a = 1.0 / n
    particle = []
    from itertools import product
    for site in product(range(n), repeat=d):
        r = a * numpy.array(site) + a/2 - 0.5
        particle.append(Particle(position=r))

    # Set spacing
    if spacing > 0:
        for p in particle:
            p.position *= spacing * n

    # Etch particles from the crystal to match the target N
    particle = random.sample(particle, N)
    return particle



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

import os
import tarfile
import numpy

from .base import TrajectoryBase
from atooms.system.particle import Particle
from atooms.system.cell import Cell
from atooms.system import System

from xml.etree import ElementTree


def map_label_id(names):
    lab = {}
    for i, key in enumerate(sorted(set(names))):
        lab[key] = i
    return lab


class TrajectoryHOOMD(TrajectoryBase):

    """HOOMD format"""

    suffix = 'tgz'

    def __init__(self, fname, mode='r'):

        super(TrajectoryHOOMD, self).__init__(fname, mode)
        base = os.path.basename(fname.split('.tgz')[0])

        if mode == 'r':
            self.__tmp_path = '/tmp/' + base
            try:
                os.makedirs(self.__tmp_path)
            except IOError:
                pass

            tar = tarfile.open(fname)
            file_list = sorted([f.name for f in tar.getmembers()])
            tar.extractall(path=self.__tmp_path)
            tar.close()
            self.__f_frames = [os.path.join(self.__tmp_path, f) for f in file_list]
            # cfg, box, pos, typ, vel = self.__read_one(self.__f_frames[0])

        elif mode == 'w':
            pass

        elif mode == 'w:gz':
            self._file = tarfile.open(fname, "w:gz")

    def read_steps(self):
        steps = []
        for f in sorted(self.__f_frames):
            tree = ElementTree.parse(f)
            root = tree.getroot()
            cfg = root.find('configuration')
            steps.append(int(cfg.attrib['time_step']))

        # First sort them, then subtract out the first step
        steps = sorted(steps)
        return [s - steps[0] for s in steps]

    def __read_one(self, fname):
        tree = ElementTree.parse(fname)
        root = tree.getroot()
        cfg = root.find('configuration')
        box = cfg.find('box')
        pos = cfg.find('position')
        try:
            vel = cfg.find('velocity')
        except:
            vel = None
        typ = cfg.find('type')
        # TODO: use tipify here
        pos_list = [list([float(x) for x in r.split()]) for r in pos.text.strip().split('\n')]
        box_list = [float(box.attrib[r]) for r in ['lx', 'ly', 'lz']]
        typ_list = typ.text.strip().split('\n')
        if vel is not None:
            vel_list = [list([float(x) for x in v.split()]) for v in vel.text.strip().split('\n')]
        else:
            vel_list = None
        return cfg, box_list, pos_list, typ_list, vel_list

    def read_system(self, frame):
        cfg, box, pos, typ, vel = self.__read_one(self.__f_frames[frame])
        if vel is None:
            particle = [Particle(species=t, position=numpy.array(p)) for p, t in zip(pos, typ)]
        else:
            particle = [Particle(species=t, position=numpy.array(p), velocity=numpy.array(v))
                        for p, t, v in zip(pos, typ, vel)]
        cell = Cell(numpy.array(box))
        return System(particle, cell)

    def write_system(self, system, step):
        ndim = len(system.particle[0].position)
        n = len(system.particle)

        if self.mode == 'w':
            fname = self.filename.split('.tgz')[0] + '_step%010d.xml' % step
        else:
            base = os.path.basename(self.filename.split('.tgz')[0])
            fname = base + '_step%010d.xml' % step

        fh = open(fname, 'w')
        fh.write("""\
<?xml version="1.0" encoding="UTF-8"?>
<hoomd_xml version="1.4">
<configuration time_step="%d" dimensions="%d" natoms="%d" >
<box lx="%g" ly="%g" lz="%g"/>
""" % ((step, ndim, n) + tuple(system.cell.side)))

        fh.write("<position num=\"%d\">\n" % n)
        for p in system.particle:
            fh.write((ndim*" %g" + "\n") % tuple(p.position))
        fh.write("</position>\n")

        fh.write("<velocity num=\"%d\">\n" % n)
        for p in system.particle:
            fh.write((ndim*" %g" + "\n") % tuple(p.velocity))
        fh.write("</velocity>\n")

        fh.write("<type num=\"%d\">\n" % n)
        for p in system.particle:
            fh.write(("%s\n") % p.species)
        fh.write("</type>\n")

        fh.write("""\
</configuration>
</hoomd_xml>
""")
        fh.close()

        if self.mode == 'w:gz':
            self._file.add(fname)
            os.remove(fname)

    def close(self):
        # clean up
        if self.mode == 'r':
            import shutil
            shutil.rmtree(self.__tmp_path)
        elif self.mode == 'w:gz':
            self._file.close()



"""Useful functions to manipulate trajectories."""

import os
import copy
import tarfile
import warnings
import subprocess
import numpy


def gopen(filename, mode):
    """Open a file recognizing gzipped and bzipped files by extension."""
    ext = os.path.splitext(filename)[1]
    if ext == '.gz':
        import gzip
        return gzip.open(filename, mode + 't')
    elif ext == '.bz2':
        import bz2
        return bz2.BZ2File(filename, mode)
    else:
        return open(filename, mode)

class _PossiblyZippedFile:

    """
    Decorator of io stream objects optimized for zipped files.

    Reading gzipped files with gzip is very slow.
    We use a simpler and faster approach, which is to unzip
    the file in a temporary directory and return that file instance.
    The temporary file is removed when closing the instance.

    Examples
    --------

    .. code-block:: python

       with ZippedFile('test.gz') as fh:
           print(fh.read())
    """

    def __init__(self, filename, mode='r'):
        self._file_tmp = None
        base, ext = os.path.splitext(filename)

        # Optimize reading
        if mode == 'r':
            # Unzip compressed files when reading
            # This is much faster than reading the gzipped object
            import tempfile
            tmp_dir = tempfile.mkdtemp()
            self._file_tmp = os.path.join(tmp_dir, os.path.basename(base))
            if ext == '.gz':
                cmd = 'gunzip'
            if ext == '.bz2':
                cmd = 'bunzip2'
            if ext in ['.gz', '.bz2']:
                subprocess.check_call(f'{cmd} -c {filename} > {self._file_tmp}', shell=True)
                self._file = open(self._file_tmp, mode)
            else:
                self._file = open(filename, mode)
        else:
            # Use gzip and bz2 modules for writing
            if ext == '.gz':
                import gzip
                self._file = gzip.open(filename, mode + 't')
            elif ext == '.bz2':
                import bz2
                self._file = bz2.BZ2File(filename, mode)
            else:
                self._file = open(filename, mode)

    def __getattr__(self, name):
        """
        Forward all unknown attribute calls to file stream
        """
        return getattr(self._file, name)

    def __enter__(self):
        return self

    # TODO: is this trapped when there is an exception?
    def __exit__(self, type, value, traceback):
        self._file.close()

    def close(self):
        from atooms.core.utils import rmd
        self._file.close()
        if self._file_tmp:
            rmd(os.path.dirname(self._file_tmp))


def file_index(fh, size=None):
    """Lightweight file indexing for trajectories via tell/seek"""
    header = []
    block = []
    block_size = []
    if size is None:
        def size(fh, data, line):
            npart = int(data)
            return 1, npart

    # Make sure we start from the beginning
    fh.seek(0)
    while True:
        line = fh.tell()
        data = fh.readline()
        # We break if file is over or we found an empty line
        if not data:
            break

        # Get size of header and block
        try:
            header_size, this_block_size = size(fh, data, line)
            header.append(line)
        except ValueError:
            raise IOError('malformed file [{}]'.format(fh.filename))

        # Skip header_size lines (if zero none will be skipped)
        for _ in range(header_size):
            fh.readline()

        # Skip block_size lines, making sure we have
        # read precisely that number of lines
        line = fh.tell()
        for _ in range(this_block_size):
            data = fh.readline()

        # Store first line /after/ we have read the frame
        # making sure the last we read was not emtpy
        # Note that readline() returns an empty string on EOF
        if len(data) > 0:
            block.append(line)
            block_size.append(this_block_size)
        else:
            raise IOError('malformed file [%s]', fh.filename)

    # TODO: leave fh as it was
    return header, block, block_size


def split(inp, path='{base}-{step:09}{ext}', archive=False, index=None):
    """
    Split the trajectory into independent trajectory files, one per
    sample.
    """
    if archive:
        tar = tarfile.open(inp.filename + '.tar.gz', "w:gz")
    base, ext = os.path.splitext(inp.filename)
    for frame, step in enumerate(inp.steps):
        args = locals()
        system = inp[frame]
        filename = path.format(**args)
        with inp.__class__(filename, 'w') as t:
            t.write(system, step)
        if archive:
            tar.add(filename)
            os.remove(filename)
    if archive:
        tar.close()


def get_block_size(data):
    """
    Return the size of the periodic block after which entries in
    `data` repeat. It is used to determine the block size in
    trajectories with logarithmic time spacing.
    """
    if len(data) < 2:
        return 1
    delta_old = 0
    delta_one = data[1] - data[0]
    iold = data[0]
    period = 1
    for ii in range(1, len(data)):
        i = data[ii]
        delta = i-iold
        # If we find that we repeat the increment between entries is
        # smaller than the previous iteration and it gets back to the
        # initial one (delat_one) then we found a block. We must
        # correct the +1 overshoot thus we subtract -1 to period
        if delta < delta_old and delta == delta_one:
            return period - 1
        else:
            period += 1
            iold = i
            delta_old = delta

    # We got to the end of the trajectory
    if len(data) != period:
        raise ValueError('something went wrong in block analysis')
    if data[1]-data[0] == data[-1]-data[-2]:
        # If the difference between steps is constant (euristically)
        # the period is one
        return 1
    else:
        # There is no periodicity, the block size is the whole trajectory
        return period


def check_block_size(steps, block_size, prune=False):
    """
    Perform some consistency checks on periodicity of non linear sampling.

    `block_size` is the number of frames composing a periodic block.
    If `prune` is True, the steps that do not match the first periodic
    block will be removed.

    Return a new list of steps that match the periodicity.

    Examples
    --------
    .. code-block:: python

       steps = [0, 1, 2, 4, 8, 9, 10, 12, 16]
       check_block_size(steps, 4)

    Note that in this case, `len(steps) % block_size == 1`, which is tolerated.
    """
    # Linear sampling
    if block_size == 1:
        return None

    # This is a single non-linear block
    if block_size == len(steps):
        return None

    steps_local = copy.copy(steps)

    # Identify steps that do not match the first periodic block
    block = steps_local[0: block_size]
    ibl, jbl = 0, 0
    prune_me = []
    for _, step in enumerate(steps_local):
        offset = block[0] if ibl > 0 else 0
        step_expected = ibl * (steps_local[block_size] - offset) + block[jbl]
        if step == step_expected:
            if jbl == block_size-1:
                # We are done with this block, we start over
                ibl += 1
                jbl = 0
            else:
                # We increment the index within the block
                jbl += 1
        else:
            prune_me.append(step)

    # Remove samples that do not conform with first block
    if prune and len(prune_me) > 0:
        for step in prune_me:
            _ = steps_local.pop(steps_local.index(step))

    # Check if the number of steps is an integer multiple of
    # block period (we tolerate a rest of 1)
    rest = len(steps_local) % block_size
    if rest > 1:
        steps_local = steps_local[:-rest]
        warnings.warn('truncated block')

    # Final test, after pruning spurious samples we should have a period
    # sampling, otherwise there was some error
    nbl = len(steps_local) // block_size
    for i in range(nbl):
        # We test that the difference between the finger print and the
        # first sample in the block is constant
        diff_last = None
        for j in range(len(block)):
            diff = steps_local[i*block_size + j] - block[j]
            if diff_last is None:
                diff_last = diff
            if diff_last != diff:
                raise IndexError('block does not match finger print {}'.format(block))
    return steps_local


def dump(trajectory, what='pos'):
    """
    Dump coordinates as a list of (npart, ndim) numpy arrays if the
    trajectory is grandcanonical or as (nsteps, npart, ndim) numpy
    array if it is not grandcanonical.
    """
    if trajectory.grandcanonical:
        data = []
        for i, s in enumerate(trajectory):
            data[i].append(s.dump(what))
    else:
        data = numpy.zeros([len(trajectory.steps),
                            len(trajectory[0].particle),
                            len(trajectory[0].cell.side)])
        for i, s in enumerate(trajectory):
            data[i] = s.dump(what)

    return data


def field(trajectory, trajectory_field, field_name, frame, x_field=None):
    """
    Return the field specified by particle attribute `field_name` at a
    given `frame`.
    """
    if x_field is not None:
        raise DeprecationWarning('use field_name instead of x_field')
        field_name = x_field
    step = trajectory.steps[frame]
    try:
        index_field = trajectory_field.steps.index(step)
    except ValueError:
        return None
    x = []
    for pi in trajectory_field[index_field].particle:
        fi = getattr(pi, field_name)
        x.append(fi)
    return x


def paste(t1, t2):
    """
    Iterate simultaneously on two trajectories. Skip samples that
    exist in one trajectory and not in the other.

    Example:
    -------
    t1 = Trajectory(f1)
    t2 = Trajectory(f2)
    for s1, s2 in paste(t1, t2):
        pass
    """
    steps_1 = set(t1.steps)
    steps_2 = set(t2.steps)
    steps = sorted(steps_1 & steps_2)
    for step in steps:
        s1 = t1[t1.steps.index(step)]
        s2 = t2[t2.steps.index(step)]
        yield step, s1, s2


def is_cell_variable(trajectory, tests=1):
    """
    Simple test to check if cell changes.

    We compare the first frame to an integer number `tests` of other
    frames starting from the end of `trajectory`.
    """
    is_variable = False
    frames = len(trajectory)
    if tests > 0:
        skip = max(1, int(frames / float(tests)))
    else:
        skip = 1
    L0 = trajectory[0].cell.side
    for sample in range(frames-1, -1, -skip):
        L1 = trajectory[sample].cell.side
        if numpy.any(L0 != L1):
            is_variable = True
            break
    return is_variable


def is_semigrandcanonical(trajectory, tests=1):
    """
    Simple test to check if a trajectory is semigrandcanonical.
    i.e. if the chemical concentrations fluctuate.

    We compare the first frame to an integer number `tests` of other
    frames starting from the end of `trajectory`.
    """
    # This is adapted from is_cell_variable()
    is_variable = False
    if tests > 0:
        skip = max(1, int(len(trajectory) / float(tests)))
    else:
        skip = 1
    from atooms.system.particle import composition
    x0 = composition(trajectory[0].particle)
    for sample in range(len(trajectory)-1, -1, -skip):
        x1 = composition(trajectory[sample].particle)
        is_variable = False
        for sp in x0:
            if x0[sp] != x1[sp]:
                is_variable = True
                break
        if is_variable:
            break
    return is_variable


def is_grandcanonical(trajectory, tests=1):
    """
    Simple test to check if a trajectory is grandcanonical.
    i.e. if the number of particles fluctuates.

    We compare the first frame to an integer number `tests` of other
    frames starting from the end of `trajectory`.
    """
    # This is adapted from is_cell_variable()
    # and basically the same code as is_semigrandcanonical()
    is_variable = False
    if tests > 0:
        skip = max(1, int(len(trajectory) / float(tests)))
    else:
        skip = 1
    N0 = len(trajectory[0].particle)
    for sample in range(len(trajectory)-1, 0, -skip):
        N1 = len(trajectory[sample].particle)
        is_variable = False
        if N0 != N1:
            is_variable = True
            break
    return is_variable


def formats():
    """Return a string with the available trajectory formats."""
    from atooms import trajectory
    txt = 'available trajectory formats:\n'
    fmts = trajectory.Trajectory.formats
    maxlen = max([len(name) for name in fmts])
    for name in sorted(fmts):
        class_name = fmts[name]
        # Check read/write capability
        try:
            fmts[name].read_system(None, None)
        except NotImplementedError:
            capability = ' '
        except:
            capability = 'R'
        else:
            capability = 'R'
        try:
            fmts[name].write_system(None, None, None)
        except NotImplementedError:
            pass
        except:
            capability += 'W'
        else:
            capability += 'W'

        # Get format description from __doc__
        if class_name.__doc__:
            docline = class_name.__doc__.strip().split('\n')[0]
            if len(docline) == 0:
                docline = '...no description...'
        else:
            docline = '...no description...'
        fmt = ' - %-' + str(maxlen) + 's : [%-2s] %s\n'
        txt += fmt % (name, capability, docline)
    return txt.strip('\n')


def info(trajectory, keys=None):
    """Return a string with information about a `trajectory` instance."""
    from atooms.system.particle import distinct_species, composition
    system = trajectory[0]
    if keys is None:

        # Default: full info
        txt = ''
        txt += 'path                 = %s\n' % trajectory.filename
        txt += 'format               = %s\n' % trajectory.__class__
        txt += 'frames               = %s\n' % len(trajectory)
        txt += 'megabytes            = %s\n' % (os.path.getsize(trajectory.filename) / 1e6)
        txt += 'particles            = %s\n' % len(system.particle)
        txt += 'species              = %s\n' % ', '.join(distinct_species(system.particle))
        txt += 'composition          = %s\n' % dict(composition(system.particle))
        txt += 'size dispersion      = {}\n'.format((numpy.std([p.radius for p in system.particle]) / numpy.mean([p.radius for p in system.particle])))
        txt += 'density              = %s\n' % round(system.density, 10)
        if system.cell is not None:
            txt += 'cell side            = %s\n' % str(list(system.cell.side))[1: -1]
            txt += 'cell volume          = %s\n' % system.cell.volume
        if len(trajectory) > 1:
            txt += 'steps                = %s\n' % trajectory.steps[-1]
            txt += 'duration             = %s\n' % trajectory.times[-1]
            txt += 'timestep             = %s\n' % trajectory.timestep
            txt += 'block size           = %s\n' % trajectory.block_size
            if trajectory.block_size == 1:
                txt += 'steps between frames = %s\n' % (trajectory.steps[1]-trajectory.steps[0])
                txt += 'time between frames  = %s\n' % (trajectory.times[1]-trajectory.times[0])
            else:
                txt += 'block steps          = %s\n' % trajectory.steps[trajectory.block_size-1]
                txt += 'block                = %s\n' % ([trajectory.steps[i] for i in range(trajectory.block_size)])
            txt += 'grandcanonical       = %s' % trajectory.grandcanonical
        return txt

    else:
        # Selected infos.
        # TODO: of course, it would be cleaner to have a little class for that
        outs = []
        for key in keys.split(','):
            if key == 'path':
                outs.append(trajectory.filename)
            elif key == 'format':
                outs.append(trajectory.__class__)
            elif key == 'frames':
                outs.append(len(trajectory))
            elif key == 'megabytes':
                outs.append(os.path.getsize(trajectory.filename) / 1e6)
            elif key == 'particles':
                outs.append(len(system.particle))
            elif key == 'species':
                outs.append(', '.join(distinct_species(system.particle)))
            elif key == 'composition':
                outs.append(dict(composition(system.particle)))
            elif key == 'cell density':
                outs.append(round(system.density, 10))
            elif key == 'cell side':
                outs.append(str(list(system.cell.side))[1: -1])
            elif key == 'cell volume':
                outs.append(system.cell.volume)
            elif key == 'steps':
                outs.append(trajectory.steps[-1])
            elif key == 'duration':
                outs.append(trajectory.times[-1])
            elif key == 'timestep':
                outs.append(trajectory.timestep)
            elif key == 'block size':
                outs.append(trajectory.block_size)
            elif key == 'steps between frames':
                outs.append(trajectory.steps[1]-trajectory.steps[0])
            elif key == 'time between frames':
                outs.append(trajectory.times[1]-trajectory.times[0])
            elif key == 'block steps':
                outs.append(trajectory.steps[trajectory.block_size-1])
            elif key == 'block':
                outs.append([trajectory.steps[i] for i in range(trajectory.block_size)])
            elif key == 'grandcanonical':
                outs.append(trajectory.grandcanonical)

        txt = ''
        fmt = '%%-%ds : %%s\n' % (max([len(key) for key in keys.split(',')]))
        for key, out in zip(keys.split(','), outs):
            txt += fmt % (key, out)

        return txt.strip('\n')



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Minimal simulation backend for LAMMPS (http://lammps.sandia.gov).
"""

import os
import subprocess
import tempfile
import numpy
from atooms import trajectory
from atooms import system
from atooms.system import interaction
from atooms.trajectory import TrajectoryLAMMPS
from atooms.trajectory.decorators import change_species
from atooms.core.utils import rmd

# Lammps command
lammps_command = 'lmp'

# MPI environment
try:
    lammps_mpi_tasks = 1
    lammps_mpi = 'mpirun'
    _ = subprocess.check_output('{lammps_mpi} --version', shell=True,
                                stderr=subprocess.STDOUT, executable='/bin/bash')
except subprocess.CalledProcessError:
    lammps_mpi_tasks = 1
    lammps_mpi = ''


def installed():
    """Return `True` if `lammps_command` can be executed"""
    try:
        _ = subprocess.check_output(f'echo | {lammps_mpi} {lammps_command}', shell=True,
                                    stderr=subprocess.STDOUT, executable='/bin/bash')
        return True
    except subprocess.CalledProcessError:
        return False


def _get_lammps_version():
    """Return lammps version and raise an exception if lammps is not installed"""
    try:
        _ = subprocess.check_output(f'echo | {lammps_mpi} {lammps_command}', shell=True,
                                    stderr=subprocess.STDOUT, executable='/bin/bash')
        version = _.decode().split('\n')[0][8:-1]
    except subprocess.CalledProcessError:
        raise ImportError(f'lammps not installed (command is {lammps_command})')
    return version

def _run_lammps_command(cmd):
    """Run a lammps script from the command line"""
    dirout = tempfile.mkdtemp()
    file_tmp = os.path.join(dirout, 'cmd.lammps')
    with open(file_tmp, 'w') as fh:
        fh.write(cmd)
    opt = f'-n {lammps_mpi_tasks}' if lammps_mpi else ''
    shell_cmd = f'{lammps_mpi} {opt} {lammps_command} -in {file_tmp}'
    try:
        stdout = subprocess.check_output(shell_cmd, shell=True,
                                         executable='/bin/bash',
                                         stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as exception:
        print(cmd)
        if isinstance(exception.output, bytes):
            print(exception.output.decode())
        else:
            print(''.join(exception.output))
        raise

    # Clean up
    rmd(dirout)
    return stdout.decode()


class Interaction(interaction.Interaction):
    """
    Interaction wrapper for LAMMPS.

    For the time being, it assumes `self.potential` is a string
    containing appropriate lammps commands that define the
    interaction.
    """
    # TODO: assign interaction to system based on pair_style entries in cmd

    def __init__(self, potential, style='atomic', units='lj'):
        interaction.Interaction.__init__(self)
        self.style = style
        self.units = units
        self.potential = potential
        self.variables = {'particle': 'particle', 'cell': 'cell'}
        self.energy, self.virial = 0.0, 0.0

    def compute(self, observable, particle, cell):
        # We use self.potential as lammps commands
        dirout = tempfile.mkdtemp()
        file_tmp = os.path.join(dirout, 'lammps.atom')
        file_inp = os.path.join(dirout, 'lammps.atom.inp')
        # Update lammps startup file using self.system
        # This will write the .inp startup file
        with TrajectoryLAMMPS(file_tmp, 'w') as th:
            th.write(system.System(particle, cell), 0)
        # Do things in lammps order: units, read, commands, run. A
        # better approach would be to parse commands and place
        # read_data after units then pack commands again. Even better
        # using PyLammps...
        cmd = f"""\
units		{self.units}
atom_style	{self.style}
read_data {file_inp}
{self.potential}
run 0
write_dump all custom {file_tmp} fx fy fz modify sort id format line "%.15g %.15g %.15g"
"""

        stdout = _run_lammps_command(cmd)

        found = False
        for line in stdout.split('\n'):
            if 'Step' in line:
                found = True
            elif found:
                if 'MPI' in line:
                    continue
                _, T, U, _, _, P = [float(x) for x in line.split()]
                rho = len(particle) / cell.volume
                ndim = len(cell.side)
                self.energy = U * len(particle)
                self.virial = (P - rho*T) * ndim * cell.volume
                break

        with TrajectoryLAMMPS(file_tmp) as th:
            new_system = th[-1]
            self.forces = new_system.interaction.forces

        # Clean up
        rmd(dirout)


# We use the base system class
System = system.System


class LAMMPS:

    """LAMMPS simulation backend."""

    def __init__(self, inp, commands, restart=False, verbose=False,
                 style='atomic', units='lj'):
        """
        We initialize the backend from `inp`, which can be a `System`, a
        `Trajectory` or path to a trajectory. LAMMPS `commands` must
        be a string or a file and should not contain dump or run
        commands.
        """
        self.version = _get_lammps_version()
        self.verbose = verbose
        self.style = style
        self.units = units

        # Actually only atomic and charge styles are supported
        assert self.style in ['atomic', 'charge']

        # Initialize commands
        self.commands = commands
        if os.path.exists(commands):
            with open(commands) as fh:
                self.commands = fh.read()

        # Define the initial system
        if isinstance(inp, system.System):
            # If we pass a system there is nothing to do
            self.system = inp

        elif isinstance(inp, trajectory.base.TrajectoryBase):
            # It is trajectory, we get the last frame
            self.system = inp[-1]

        elif os.path.exists(inp):
            # We accept any trajectory format, but if the format is
            # not recognized we force lammps native (atom) format
            try:
                with trajectory.Trajectory(inp) as t:
                    # We enforce fortran species layout
                    t.add_callback(change_species, 'F')
                    s = t[-1]
            except ValueError:
                with trajectory.TrajectoryLAMMPS(inp) as t:
                    s = t[-1]
            self.system = s

        else:
            raise ValueError(f'could not initialize system from {inp}')

        # Default trajectory format
        self.trajectory_class = TrajectoryLAMMPS

        # Assign commands as potentials
        # TODO: strip commands unrelated to interaction
        # TODO: if the interaction is already present, there should be no interaction info in commands
        if self.system.interaction is None:
            self.system.interaction = Interaction(commands, style=style, units=units)

        # Initialize the indices of the periodic cell the particle is in
        for p in self.system.particle:
            p._ix, p._iy, p._iz = 0, 0, 0

        # Tmp directory for restart files
        # TODO: clean this somehow
        self.restart = restart
        if self.restart:
            self.tmpdir = tempfile.mkdtemp()

        # Set unfolded position and store a copy of the initial system
        for p in self.system.particle:
            if hasattr(p, 'position_unfolded'):
                p.position_unfolded[:] = p.position.copy()
            else:
                p.position_unfolded = p.position.copy()
        self.initial = self.system.__class__()
        self.initial.update(self.system, exclude=['interaction'])

    def __str__(self):
        return 'LAMMPS'

    # TODO: refactor this via Dynamics class borrowed from atooms.dynamics
    # There should be a check then in partial_rmsd on the array order
    # Note: the views here are irrelevant because the particle objects are
    # reinstatiated after every run
    @property
    def rmsd(self):
        current = self.system.view('particle.position_unfolded')
        initial = self.initial.view('particle.position_unfolded')
        msd = numpy.sum((current - initial)**2) / len(self.system.particle)
        return msd**0.5

    @property
    def partial_rmsd(self):
        current = self.system.view('particle.position_unfolded')
        initial = self.initial.view('particle.position_unfolded')
        initial_species = self.system.view('particle.species')
        rmsd = []
        # Note: assume C-layout for lammps backend
        for species in self.system.distinct_species:
            mask = initial_species == species
            msd = numpy.sum((current[mask, :] - initial[mask, :])**2) / len(mask)
            rmsd.append(msd**0.5)
        return rmsd

    def read_checkpoint(self, output_path):
        pass

    def write_checkpoint(self, output_path):
        pass

    def run(self, steps):
        dirout = tempfile.mkdtemp()
        file_tmp = os.path.join(dirout, 'lammps.atom')
        file_inp = os.path.join(dirout, 'lammps.atom.inp')
        if self.restart:
            file_res = os.path.join(self.tmpdir, 'lammps.restart')

        # Update lammps startup file using self.system
        # This will write the .inp startup file
        with TrajectoryLAMMPS(file_tmp, 'w') as th:
            th.write(self.system, 0)

        # Set fixes from the system if we find thermostat / barostat
        if 'fix' not in self.commands:
            if self.system.thermostat is not None and self.system.barostat is not None:
                # NPT ensemble
                fix = 'fix 1 all npt temp {0.temperature} {0.temperature} {0.relaxation_time} iso {1.pressure} {1.pressure} {1.relaxation_time}'.format(
                    self.system.thermostat, self.system.barostat)

            elif self.system.thermostat is not None:
                # NVT ensemble
                if self.system.number_of_dimensions == 3:
                    fix = 'fix 1 all nvt temp {0.temperature} {0.temperature} {0.relaxation_time}'.format(self.system.thermostat)
                else:
                    fix = """
compute myTemp all temp/partial 1 1 0
fix 1 all nvt temp {0.temperature} {0.temperature} {0.relaxation_time}
fix_modify 1 temp myTemp
""".format(self.system.thermostat)

            else:
                # NVE ensemble
                fix = 'fix 1 all nve'
        else:
            # The integrator is already contained in the commands
            assert self.system.thermostat is None and self.system.barostat is None, \
                'lammps script contains a fix already'
            fix = ''

        # Do things in lammps order: units, read, commands, run. A
        # better approach would be to parse commands and place
        # read_data after units then pack commands again. Even better
        # using PyLammps...
        cmd = f"""\
units		{self.units}
atom_style	{self.style}
"""
        # Read restart file if it exists
        if self.restart and os.path.exists(file_res):
            cmd += f"""
read_restart {file_res}
"""
        else:
            cmd += f"""
read_data {file_inp}
"""

        # Rest of commands
        cmd += f"""
{self.commands}
{fix}
run {steps}
"""

        # Store system state at the end of run to update the System instance
        _dump_args = {'atomic': 'id type mass x y z vx vy vz ix iy iz',
                      'charge': 'id type mass q x y z vx vy vz ix iy iz'}
        _dump_fmt = {'atomic': '%d %d %.15g %.15g %.15g %.15g %.15g %.15g %.15g %d %d %d',
                     'charge': '%d %d %.15g %.15g %.15g %.15g %.15g %.15g %.15g %.15g %d %d %d'}

        dump_args = _dump_args[self.style]
        dump_fmt = _dump_fmt[self.style]
        cmd += f"""
write_dump all custom {file_tmp} {dump_args} modify sort id format line "{dump_fmt}"
"""

        if self.restart:
            cmd += f"""
write_restart {file_res}
"""

        # Execute LAMMPS command
        stdout = _run_lammps_command(cmd)
        if self.verbose:
            print(stdout)

        # Update internal reference to self.system
        # Note that the thermostat and barostat are not touched
        def _fix_2d(system):
            system.cell.side = system.cell.side[:2]
            for p in system.particle:
                p.position = p.position[:2]
                p.velocity = p.velocity[:2]
            return system

        with TrajectoryLAMMPS(file_tmp) as th:
            if self.system.number_of_dimensions == 2:
                th.add_callback(_fix_2d)
            new_system = th[-1]

        # Note: the particles' objects get reset here, views will be broken!
        # TODO: can we really update the particle instances to preserve views?
        self.system.update(new_system, exclude=['thermostat', 'barostat',
                                                'interaction'])
        if self.system.number_of_dimensions == 3:
            for p in self.system.particle:
                p.position_unfolded = p.position + numpy.array([p._ix, p._iy, p._iz]) * new_system.cell.side
        if self.system.number_of_dimensions == 2:
            for p in self.system.particle:
                p.position_unfolded = p.position + numpy.array([p._ix, p._iy]) * new_system.cell.side

        # Clean up
        rmd(dirout)


class EnergyMinimization(LAMMPS):

    """LAMMPS minimization backend."""

    def __init__(self, inp, commands, method='cg', ftol=1e-4, steps=100000):
        """
        We initialize the backend from `inp`, which can be a `System`, a
        `Trajectory` or path to a trajectory. LAMMPS `commands` must
        be a string or a file and should not contain dump or minimize
        commands.
        """
        LAMMPS.__init__(self, inp, commands)
        self.steps = steps
        self.ftol = ftol
        self.method = method
        self.max_evaluations = 100000
        # Optimization backends must set a boolean reached_steps
        # attribute. It is True at the beginning.
        self.reached_steps = True

    def __str__(self):
        return 'LAMMPS energy minimization'

    def run(self, steps=None):
        if steps is not None:
            self.steps = steps

        dirout = tempfile.mkdtemp()
        file_tmp = os.path.join(dirout, 'lammps.atom')
        file_inp = os.path.join(dirout, 'lammps.atom.inp')

        # Update lammps startup file using self.system
        # This will write the .inp startup file
        with TrajectoryLAMMPS(file_tmp, 'w') as th:
            th.write(self.system, 0)

        # Do things in lammps order: units, read, commands, run. A
        # better approach would be to parse commands and place
        # read_data after units then pack commands again. Even better
        # using PyLammps...
        max_evaluations = min(100000000, self.max_evaluations)
        cmd = f"""\
units		{self.units}
atom_style	{self.style}
read_data {file_inp}
{self.commands}
min_style {self.method}
minimize 0.0 {self.ftol} {self.steps} {max_evaluations}
write_dump all custom {file_tmp} id type x y z modify sort id format line "%d %d %.15g %.15g %.15g"
"""
        stdout = _run_lammps_command(cmd)

        if self.verbose:
            print(stdout)

        # Check if we have reached the number of maximum number of steps
        if 'Stopping criterion = max iterations' in stdout:
            self.reached_steps = True
        else:
            self.reached_steps = False

        # Update internal reference to self.system
        with TrajectoryLAMMPS(file_tmp) as th:
            new_system = th[-1]
        for i in range(len(self.system.particle)):
            self.system.particle[i] = new_system.particle[i]

        # Clean up
        rmd(dirout)



import logging
import numpy
import f2py_jit
from .linked_cells import LinkedCells
from .helpers import _merge_source

_log = logging.getLogger(__name__)


class VerletList:

    def __init__(self, skin=0.2, update="largest", update_period=10, helpers='helpers.f90',
                 inline=True, debug=False, parallel=False, binning="auto"):
        self.skin = skin
        self.update = update
        self.update_period = update_period
        self.displacement = None
        self.rcut = None
        self.number_of_neighbors = None
        self.neighbors = None
        self.is_adjusted = False
        self.calls = 0
        self._last_call = 0
        self.binning = binning
        self._linked_cells = None
        self._boost = 1.0

        # Merge all sources into a unique source blob
        source = _merge_source(helpers, 'neighbor_list.f90')

        if inline:
            from f2py_jit.finline import inline_source
            source = inline_source(source)

        # TODO: refactor into a helper function
        # with open('/tmp/dump.f90', 'w') as fh:
        #     fh.write(source)
        args, opt_args = '', '-O3 -ffree-form -ffree-line-length-none'
        if debug:
            opt_args += ' -pg -fbounds-check'
        if parallel:
            opt_args += ' -fopenmp'
            args += ' -lgomp'
        extra_args = '--opt="{}" {}'.format(opt_args, args)
        self._uid = f2py_jit.build_module(source, extra_args=extra_args,
                                          metadata={'neighbors': 'neighbor_list.f90',
                                                    'helpers': helpers,
                                                    'inline': inline,
                                                    'parallel': parallel},
                                          )
        # This way we cannot deepcopy but a bit more efficient (because importing from cache locks...)
        self._f90 = f2py_jit.import_module(self._uid)

    def __str__(self):
        return f'Verlet list (skin={self.skin}, update={self.update})'

    def adjust(self, box, pos, rcut=None, boost=1.0, force=False):
        """
        Adjust neighbor list to positions of particles in the box and cut off distance.
        """
        from math import gamma, pi
        assert boost == 1.0, 'boost is deprecated'
        assert rcut is not None or force, 'provide rcut'

        # We return immediately is the list has been adjusted already
        # and N has not changed or we are not forcing a readjustement
        if self.is_adjusted and (self.displacement.shape == pos.shape and not force):
            return

        # If we provide rcut, we store it so that we can automatically adjust the list
        # TODO: in the unlikely event that rcut changes inplace, we should copy it
        if rcut is not None:
            self._rcut = rcut
            self.rcut = rcut + self.skin
        else:
            self.rcut = self._rcut + self.skin

        # Adjust binning is necessary
        if self.binning == "auto" or self.binning:
            self._linked_cells = LinkedCells(numpy.min(self.rcut), newton=True)
            self._linked_cells._adjust(box)
            self._linked_cells._is_adjusted = True
            if max(self._linked_cells.n_cell) <= 3:
                self._linked_cells = None

        # Set up neighbor list
        ndim, N = pos.shape
        volume = pi**(ndim/2.) / gamma(ndim/2. + 1) * numpy.max(self.rcut)**ndim
        rho = N / numpy.prod(box)
        nmax = max(1, int(rho * volume * 1.5 * self._boost))
        self.displacement = numpy.zeros_like(pos, order='F')
        self.number_of_neighbors = numpy.zeros(N, dtype=numpy.int32, order='F')
        self.neighbors = numpy.zeros((nmax, N), dtype=numpy.int32, order='F')
        self._distances = numpy.zeros((nmax, N), dtype=float, order='F')
        self._pos_old = pos.copy(order='F')
        self.is_adjusted = True

    def add_displacement(self, pos, box):
        """
        Accumulate displacements assuming that PBC has been applied, hence we need to refold them.
        """
        self._f90.neighbor_list.add_displacement(pos, self._pos_old, box, self.displacement)

    def need_update(self):
        """Return True if neighbor list must be completely updated."""
        if self.update == 'periodic':
            return self.calls % self.update_period == 0
        if self.update == 'largest':
            return self._f90.neighbor_list.need_update_largest(self.displacement, self.skin)
        else:
            raise ValueError(f'unknown update method {self.update}')

    def compute(self, box, pos, ids, force_update=False):
        """Compute Verlet lists"""
        self.add_displacement(pos, box)
        if self.calls == 0 or self.need_update() or force_update:
            if self._linked_cells:
                self._linked_cells.compute(box, pos, ids)
                error = self._f90.neighbor_list.compute_from_neighbors(box, pos, ids, self.rcut,
                                                                       self._linked_cells.neighbors + 1,
                                                                       self._linked_cells.number_neighbors,
                                                                       self.neighbors, self.number_of_neighbors,
                                                                       self._distances)
            else:
                error = self._f90.neighbor_list.compute(box, pos, ids, self.rcut,
                                                        self.neighbors, self.number_of_neighbors, self._distances)
            if error:
                self._boost *= 1.1
                self.adjust(box, pos, force=True)
                _log.info('increasing neighbor list size to {} with {}x boost'.format(self.neighbors.shape[0], self._boost))
                self.compute(box, pos, ids, force_update=True)

            # This parallel piece does not help
            # self._f90.neighbor_list.zero(self.displacement)
            self.displacement[:, :] = 0.
            _log.debug('updated list after {} calls'.format(self.calls - self._last_call))
            self._last_call = self.calls

        self.calls += 1



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Base optimization class.

`atooms` provides a generic optimization interface that abstracts out
most of the common parts of particle-based optimizations.

The actual optimization code is wrapped by a simulation "backend" that
exposes a minimal but coherent interface.

`Optimization` uses callbacks to analyze and process optimization data
on the fly. The interval in steps at which callbacks are executed is
controlled by a `Scheduler` instance.
"""

import logging

from atooms.simulation import Simulation
from atooms.simulation.observers import SimulationEnd

_log = logging.getLogger(__name__)


class OptimizationEnd(SimulationEnd):
    """Raised when an targeter reaches its target."""
    pass

def _target_backend_tolerance(sim, value):
    if not sim.backend.reached_steps:
        raise OptimizationEnd('reached backend tolerance')

def _target_force_norm_square(sim, value):
    if sim.system.force_norm_square(per_particle=True) < value:
        raise OptimizationEnd('reached target W %d' % value)


class Optimization(Simulation):

    """Optimization base class."""

    def __init__(self, backend, output_path=None, tolerance=None,
                 steps=1e6):
        """
        Perform an optimization using the specified `backend` and
        optionally write output to `output_path`. This can be a file
        or directory path.

        Paths: to define output paths we rely on `output_path`, all
        other paths are defined based on it and on its base_path.
        """
        Simulation.__init__(self, backend, output_path=output_path)
        self.steps = int(steps)
        self.tolerance = tolerance
        self._check_interval = 1000

    def __str__(self):
        return 'atooms optimization via %s' % self.backend

    def write_checkpoint(self):
        """Write checkpoint to allow restarting a simulation."""
        pass

    def read_checkpoint(self):
        """Read the checkpoint to restart a simulation."""
        pass

    def run(self, steps=None):
        if self.tolerance is not None:
            # We need to add this check only if targeters are present
            self.add(_target_backend_tolerance, self._check_interval, 0.0)
            self.add(_target_force_norm_square, self._check_interval, self.tolerance)
        Simulation.run(self, steps)



#!/usr/bin/env python

import unittest

import numpy
from atooms.system import System, Particle, Cell
from atooms.trajectory.ram import TrajectoryRamFull
from atooms.trajectory.ram import TrajectoryRamLight as TrajectoryRam

class Test(unittest.TestCase):

    def setUp(self):
        pass

    def test_ram_inplace(self):
        particle = [Particle(position=[0.0, 0.0, 0.0])]
        system = System(particle)
        t = TrajectoryRam()
        t[0] = system
        particle[0].position += 1.0
        self.assertFalse((t[0].particle[0].position == particle[0].position).all())

    def test_ram(self):
        particle = [Particle(position=[0.0, 0.0, 0.0])]
        system = System(particle)
        t = TrajectoryRam()
        t[0] = system
        particle[0].position = numpy.array([1.0, 1.0, 1.0])
        self.assertFalse((t[0].particle[0].position == particle[0].position).all())

    def test_ram_full(self):
        particle = [Particle(position=[0.0, 0.0, 0.0])]
        system = System(particle)
        t = TrajectoryRamFull()
        t[0] = system
        system.particle[0].position = numpy.array([2.0, 2.0, 2.0])
        t[0] = system
        particle[0].position += 1.0
        self.assertFalse((t[0].particle[0].position == particle[0].position).all())

    def test_ram_lammps(self):
        import os
        import numpy
        import sys
        from atooms.backends.lammps import LAMMPS
        from atooms.simulation import Simulation
        import atooms.backends.lammps
        if not atooms.backends.lammps.installed():
            self.skipTest('missing LAMMPS')
        input_file = os.path.join(os.path.dirname(__file__),
                                  '../data/lj_N1000_rho1.0.xyz')
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        fix             1 all nve
        """
        bck = LAMMPS(input_file, cmd)
        sim = Simulation(bck)
        sim.system.particle[0].position = numpy.zeros(3)

        t = TrajectoryRamFull()
        t[0] = sim.system
        sim.system.particle[0].position += 1.0
        self.assertFalse((sim.system.particle[0].position == t[0].particle[0].position).all())

    def test_ram_unfolded(self):
        import os
        import numpy
        import random
        from atooms.backends.lammps import LAMMPS
        from atooms.trajectory import TrajectoryXYZ, Unfolded
        from atooms.simulation import Simulation
        from atooms.system import Thermostat
        import atooms.backends.lammps
        if not atooms.backends.lammps.installed():
            self.skipTest('missing LAMMPS')

        def store(sim, ram):
            ram.write(sim.system, step=sim.current_step)

        tf = TrajectoryRamFull()
        t = TrajectoryRam()
        with TrajectoryXYZ(os.path.join(os.path.dirname(__file__),
                                        '../data/ka_N150_rho1.200.xyz')) as tinp:
            for i, s in enumerate(tinp):
                tf[i] = s
                t[i] = s
        tuf = Unfolded(tf, fixed_cm=True)
        tu = Unfolded(t, fixed_cm=True)
        for s, sf in zip(tu, tuf):
            for p, pf in zip(s.particle, sf.particle):
                self.assertTrue((p.position == pf.position).all())

        def mobility(t):
            import numpy as np
            side = t[0].cell.side
            K = 0
            unfoldedtj = Unfolded(t, fixed_cm=True)
            pos_0 = unfoldedtj[0].dump('pos')
            for j in range(1, len(t)):
                pos_1 = unfoldedtj[j].dump('pos')
                K += np.sum((pos_1 - pos_0)**2)
                pos_0 = pos_1
            return K

        # print mobility(t), mobility(tf)

    def test_ram_copy(self):
        particle = [Particle(position=[0.0, 0.0, 0.0])]
        system = System(particle)
        t = TrajectoryRam()
        t[0] = system
        t[0].particle[0].position = numpy.array([1.0, 1.0, 1.0])
        # print system.particle[0].position, t[0].particle[0].position
        # print id(t[0].particle[0])
        # print id(t[0].particle[0])

        particle = [Particle(position=[0.0, 0.0, 0.0])]
        system = System(particle)
        s = System(particle)
        t = TrajectoryRamFull()
        t[0] = system
        s.update(t[0])
        s.particle[0].position = numpy.array([1.0, 1.0, 1.0])
        # print system.particle[0].position, t[0].particle[0].position, s.particle[0].position
        # print id(t[0].particle[0])
        # print id(t[0].particle[0])

        particle = [Particle(position=[0.0, 0.0, 0.0])]
        system = System(particle)
        t = TrajectoryRamFull()
        t[0] = system
        system.particle[0].position = numpy.array([1.0, 1.0, 1.0])
        # print system.particle[0].position, t[0].particle[0].position
        # print id(t[0].particle[0])
        # print id(system.particle[0])

        # self.assertFalse((t[0].particle[0].position == particle[0].position).all())

    def test_ram_lammps_write(self):
        import os
        import numpy
        import sys
        from atooms.backends.lammps import LAMMPS
        from atooms.simulation import Simulation
        from atooms.simulation.observers import write_to_ram
        import atooms.backends.lammps
        if not atooms.backends.lammps.installed():
            self.skipTest('missing LAMMPS')

        input_file = os.path.join(os.path.dirname(__file__),
                                  '../data/lj_N1000_rho1.0.xyz')
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        fix             1 all nve
        """
        bck = LAMMPS(input_file, cmd)
        ram = TrajectoryRamFull()
        sim = Simulation(bck)
        sim.add(write_to_ram, 10, ram)
        sim.run(100)
        self.assertEqual(len(ram.steps), 11)

    def test_callback_copy(self):
        import copy
        from atooms.trajectory.decorators import filter_species
        particle = [Particle(species='A'), Particle(species='B')]
        system = System(particle)
        t = TrajectoryRamFull()
        t[0] = system
        t.add_callback(copy.deepcopy)
        t.add_callback(filter_species, 'A')
        self.assertEqual(t[0].distinct_species, ['A'])
        t.callbacks.pop()
        t.callbacks.pop()
        self.assertEqual(t[0].distinct_species, ['A', 'B'])


if __name__ == '__main__':
    unittest.main()



import os

def _normalize_path(path, search_dir=None):
    if not path.endswith('.f90'):
        path = path + '.f90'

    if os.path.exists(path):
        return path
    else:
        if search_dir is None:
            search_dir = os.path.dirname(__file__)
        full_path = os.path.join(search_dir, path)
        if os.path.exists(full_path):
            return full_path
        else:
            raise ValueError('could not find source for {}'.format(path))


def _merge_source(*paths):
    """Merges `sources` into a unique source."""
    merged_src = ''
    for path in paths:
        # Check path existence
        source_path = _normalize_path(path)
        with open(source_path) as fh:
            src = fh.read()
        # Merge sources into a single one
        merged_src += src
    return merged_src

def _check_potential_derivatives(potential, params, r):
    import numpy
    import f2py_jit

    u, w, h = numpy.array(0.0), numpy.array(0.0), numpy.array(0.0)
    u1, w1, h1 = numpy.array(0.0), numpy.array(0.0), numpy.array(0.0)
    u2, w2, h2 = numpy.array(0.0), numpy.array(0.0), numpy.array(0.0)
    path = _normalize_path(potential)
    f90 = f2py_jit.jit(path)
    f90.potential.init(*params)
    f90.potential.compute(1, 1, r**2, u, w, h)

    fail_w, fail_h = True, True
    dr = 1e-2
    tol = 1e-6
    while (fail_w or fail_h) and dr > 1e-14:
        f90.potential.compute(1, 1, (r+dr)**2, u1, w1, h1)
        f90.potential.compute(1, 1, (r-dr)**2, u2, w2, h2)
        w_approx = - (u1 - u2) / (2*dr) / r
        h_approx = - (w1 - w2) / (2*dr) / r
        if fail_w and abs(w - w_approx) < tol:
            fail_w = False
        if fail_h and abs(h - h_approx) < tol:
            fail_h = False
        dr /= 2
    # print('Checks for potential {} pass: {}'.format(potential, not fail_w or not fail_h))
    if fail_w:
        print(w, w_approx, abs(w - w_approx), '# w failed')
        return False
    if fail_h:
        print(w, w_approx, abs(h - h_approx), '# h failed')
        return False
    return True



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Base simulation class with callback logic.

`atooms` provides a generic simulation interface that abstracts out
most of the common parts of particle-based simulations.

`Simulation` uses callbacks to analyze and process simulation data on
the fly. The module `atooms.simulation.observers` provides basic
callbacks to write data to disk, e.g. trajejectory files, and to stop
the simulation when certain targets are reached, e.g. mean squared
displacement larger than a threshold.

The interval in steps at which callbacks are executed is controlled by
a `Scheduler` instance.

The actual simulation code is wrapped by a simulation "backend" that
exposes a minimal but coherent interface.
"""

import os
import time
import datetime
import logging

import atooms.core.progress
from atooms.core import __version__
from atooms.core.utils import mkdir, barrier
from atooms.trajectory import TrajectoryXYZ
from .observers import target_steps, Speedometer, Scheduler, SimulationEnd, SimulationKill

_log = logging.getLogger(__name__)


def _report(info, file_handle=None, log_echo=True):
    """
    Log `info` string to default logger at level info.

    Optionally write `info` to `file_handle` if the latter is
    given. Logging is disabled via `log_echo` is False.
    """
    if info is None:
        return

    if log_echo:
        for line in info.split('\n'):
            _log.info(line.strip())

    if file_handle is not None:
        file_handle.write(info)


def _callable_name(callback):
    try:
        # This is a function
        name = callback.__name__
    except AttributeError:
        # This is callable class
        name = callback.__class__.__name__
    return name.lower()


class Simulation:

    """Simulation base class."""

    version = __version__

    def __init__(self, backend, output_path=None, steps=0,
                 checkpoint_interval=0, enable_speedometer=False,
                 restart=False):
        """
        Perform a simulation using the specified `backend` and optionally
        write output to `output_path`. This can be a file or directory path.

        Paths: to define output paths we rely on `output_path`, all
        other paths are defined based on it and on its base_path.
        """
        self.backend = backend
        self.output_path = output_path
        self.steps = steps
        self._restart = restart
        self.current_step = 0
        self.initial_step = 0
        # We expect subclasses to keep a ref to the trajectory class
        # used to store configurations
        if hasattr(self.backend, 'trajectory_class'):
            self.trajectory_class = self.backend.trajectory_class
        else:
            self.trajectory_class = TrajectoryXYZ
        # Make sure the dirname of output_path exists. For instance,
        # if output_path is data/trajectory.xyz, then data/ should
        # exist. This creates the data/ folder and its parents folders.
        if self.output_path is not None:
            mkdir(os.path.dirname(self.output_path))

        # Internal variables
        self._observer = []
        self._observer_id = 0
        self._start_time = -1.0
        self._stop_time = -1.0
        self._speedometer = None
        self._checkpoint_scheduler = Scheduler(checkpoint_interval)
        self.checkpoint_variables = ['species', 'position', 'velocity', 'radius']
        self._targeter_steps = target_steps
        if enable_speedometer:
            self._speedometer = Speedometer()
            self.add(self._speedometer, Scheduler(self.steps, calls=20))

    @property
    def system(self):
        # Note that setting system as a reference in the instance, like
        #   self.system = self.backend.system
        # is unsafe because this won't follow the backend's system when the
        # latter is reassigned as in
        #   self.backend.system = None
        # So we defined it as a property.
        return self.backend.system

    @system.setter
    def system(self, value):
        self.backend.system = value

    def __str__(self):
        return 'atooms simulation via %s' % self.backend

    @property
    def restart(self):
        """True is the simulation should be restarted. Read-only property."""
        return self._restart

    @property
    def base_path(self):
        return os.path.splitext(self.output_path)[0]

    def add_callback(self, callback, scheduler, *args, **kwargs):
        self.add(callback, scheduler, *args, **kwargs)

    def add(self, callback, scheduler, *args, **kwargs):
        """
        Add an observer `callback` to be called along with a `scheduler`.

        `scheduler` and `callback` must be callables accepting a
        Simulation instance as unique argument. `scheduler` must
        return the next step at which the observer has to be notified.

        An integer value is allowed for `scheduler`. In this case, a
        scheduler with fixed interval is generated internally and the
        observer is notified every `scheduler` steps.
        """
        # Accept an integer interval
        if type(scheduler) is int:
            scheduler = Scheduler(scheduler)

        # Sanity check (prevent scheduler/callback inversion)
        assert type(scheduler(self)) is int, \
            "probable swap between callback {} and scheduler {}".format(callback, scheduler)

        # Store callback id, scheduler, callback and its arguments
        self._observer_id += 1
        pack = {'id': self._observer_id,
                'scheduler': scheduler,
                'callback': callback,
                'args': args,
                'kwargs': kwargs}

        # Keep targeters last
        if 'target' not in _callable_name(callback):
            self._observer.insert(0, pack)
        else:
            self._observer.append(pack)
        return self._observer_id

    def remove(self, callback):
        """Remove a callback."""
        if isinstance(callback, int):
            for cbk in self._observer:
                if cbk['id'] == callback:
                    self._observer.remove(cbk)
                    break
        else:
            # This is a backward compatible fix to remove a callback as function
            # It may be removed from v.4 onwards
            for cbk in self._observer:
                # Here we pretend callback_id is a function
                # This will remove all instances
                if cbk['callback'] == callback:
                    self._observer.remove(cbk)

    def remove_callback(self, callback):
        self.remove(callback)

    def _notify(self, observers):
        for observer in observers:
            _log.debug('notify %s at step %d', observer, self.current_step)
            callback = observer['callback']
            args = observer['args']
            kwargs = observer['kwargs']
            callback(self, *args, **kwargs)

    @property
    def _targeters(self):
        return [o for o in self._observer if 'target' in _callable_name(o['callback'])]

    @property
    def _non_targeters(self):
        return [o for o in self._observer if 'target' not in _callable_name(o['callback'])]

    @property
    def _speedometers(self):
        return [o for o in self._observer if isinstance(o['callback'], Speedometer)]

    def write_checkpoint(self):
        """Write checkpoint to allow restarting a simulation."""
        if self.output_path is None:
            return

        # We clear the checkpoint first.
        # This way, if we get to write the step at the end of this function
        # we are sure the checkpoint is consistent
        from atooms.core.utils import rmf
        rmf(self.output_path + '.chk.step')

        if hasattr(self.backend, 'write_checkpoint'):
            # Use native backend checkpoint method
            self.backend.write_checkpoint(self.output_path)
        else:
            # TODO: use pickle.dump
            # Fallback to backend trajectory class with high precision
            with self.trajectory_class(self.output_path + '.chk', 'w') as t:
                t.variables = self.checkpoint_variables
                t.precision = 12
                t.write(self.system)

        # Checkpoint number of steps
        with open(self.output_path + '.chk.step', 'w') as fh:
            fh.write('%d' % self.current_step)

    def read_checkpoint(self):
        """
        Read the checkpoint to restart a simulation.

        If the checkpoint file is not found, this method exits
        gracefully.
        """
        if self.output_path is None:
            return

        if os.path.exists(self.output_path + '.chk.step'):
            with open(self.output_path + '.chk.step') as fh:
                self.current_step = int(fh.read())
            _log.info('restart simulation from step {}'.format(self.current_step))
        else:
            _log.debug('could not find steps file for checkpoint')
            # If the default checkpoint file is found,
            # this means its writing was incomplete
            if os.path.exists(self.output_path + '.chk'):
                _log.error('checkpoint cannot be read, most likely in a broken state')
                raise OSError('error while reading checkpoint file {}'.format(self.output_path + '.chk'))
            return

        if hasattr(self.backend, 'read_checkpoint'):
            # Use native backend checkpoint method
            self.backend.read_checkpoint(self.output_path)
        else:
            # Fallback to backend trajectory class with high precision
            # TODO: use pickle.load
            if os.path.exists(self.output_path + '.chk'):
                with self.trajectory_class(self.output_path + '.chk') as t:
                    # Trajectory may not store the interaction,
                    # thermostat, barostat, so we must preserve it
                    self.system.update(t[0])

    @property
    def rmsd(self):
        if hasattr(self.backend, 'rmsd'):
            return self.backend.rmsd
        else:
            return 0.0

    def _elapsed_wall_time(self):
        """Elapsed wall time in seconds."""
        if self._stop_time < 0 and self._start_time > 0:
            return time.time() - self._start_time
        else:
            return self._stop_time - self._start_time

    def wall_time(self, per_step=False, per_particle=False):
        """
        Elapsed wall time in seconds.

        Optionally normalized per step and or per particle. It can be
        subclassed by more complex simulation classes.
        """
        norm = 1.0
        # Normalize per particle
        if per_particle:
            if len(self.system.particle) > 0:
                norm *= len(self.system.particle)
            else:
                return float('nan')
        # Normalize per step
        if per_step:
            if self.current_step - self.initial_step > 0:
                norm *= (self.current_step - self.initial_step)
            else:
                return float('nan')
        return self._elapsed_wall_time() / norm

    def run_until(self, steps):
        """
        Run the simulation up to `steps`.

        Subclasses must set steps.
        """
        self.backend.run(steps - self.current_step)
        self.current_step = steps

    def run(self, steps=None):
        """Run the simulation."""
        if steps is None:
            steps = self.steps
        else:
            if self.restart and self.current_step > 0:
                # If we are restarting, we do not allow changing
                # target steps on the fly, because it might have side
                # effects like non-constant writing interval.
                _log.warning('attempt to change steps while restarting')
                steps = self.steps
            else:
                # Keep track of the number of steps
                self.steps = steps

        # Make sure we have an integer number of steps
        steps = int(round(steps))

        # Targeter for steps. We first replace the existing one
        self.remove(self._targeter_steps)
        self.add(self._targeter_steps, Scheduler(self.current_step + steps),
                 self.current_step + steps)

        # Targeter for progress bar.
        if atooms.core.progress.active:
            # Add a dummy callback to trigger update of progress bar
            # when observers have a long interval. We try to flush the
            # progress bar 10 times.
            intervals = [cbk['scheduler'].interval for cbk in self._observer]
            intervals = [interval for interval in intervals if interval is not None]
            min_iters = 10
            if len(intervals) == 0 or (min(intervals) > (self.current_step + steps) / min_iters and
                                       (self.current_step + steps) / min_iters > 10):
                def flush(sim):
                    pass
                self.add(flush, Scheduler((self.current_step + steps) // min_iters))

        # Report
        _report(self._info_start())
        _report(self._info_backend())
        _report(self._info_observers())
        try:
            _report(str(self.system))
        except:
            _log.warning('issues reporting system; this is mostly harmless but check client code')
        _report(str(self.backend))

        # Read checkpoint if we restart
        if self.restart:
            self.read_checkpoint()
        barrier()

        # Add SIGTERM handler
        import signal

        def signal_term_handler(signal, frame):
            raise SimulationKill('simulation terminated')

        signal.signal(signal.SIGTERM, signal_term_handler)

        # Reinitialize speedometers
        for s in self._speedometers:
            s['callback']._init = False
        from atooms.core.progress import progress
        bar = progress(total=self.steps)

        # Ready to go
        self.initial_step = self.current_step
        # Timing can be refactored with core.utils.Timer()
        # Restart timing if the simulation is started again
        if self._stop_time > 0:
            self._start_time = self._stop_time
        else:
            self._start_time = time.time()
        self._stop_time = -1
        try:
            # Before entering the simulation, check if we can quit right away
            # TODO: this should be moved outside this block to avoid rewriting checkpoint / logs
            self._notify(self._targeters)

            # If this is not the first step it means we are restarting
            # and these observers have been already called at the end
            # of the last checkpointed step. So we do not need to call
            # them again.
            if self.current_step == 0:
                self._notify(self._non_targeters)

            while True:
                # Run simulation until any of the observers need to be called
                # TODO: change naming
                all_steps = [c['scheduler'](self) for c in self._observer]
                next_checkpoint = self._checkpoint_scheduler(self)
                next_step = min(all_steps + [next_checkpoint])

                self.run_until(next_step)

                # Find observers indexes corresponding to minimum step
                # then get all corresponding observers
                next_step_ids = [i for i, step in enumerate(all_steps) if step == next_step]
                next_observers = [self._observer[i] for i in next_step_ids]

                # Observers should be sorted such that targeters are
                # last to avoid cropping output files
                self._notify(next_observers)
                if self.current_step == next_checkpoint:
                    self.write_checkpoint()

                # Update progress bar
                bar.update(self.current_step)

        except SimulationEnd as end:
            # Checkpoint configuration at last step
            bar.update(self.current_step)
            bar.close()
            _log.info('simulation ended successfully: %s', end)
            self.write_checkpoint()
            _report(self._info_end())

        except KeyboardInterrupt:
            pass

        except SimulationKill:
            _log.info('simulation terminated')

        except:
            _log.error('simulation failed')
            raise

        finally:
            self._stop_time = time.time()

    def _info_start(self):
        now = datetime.datetime.now().strftime('%Y-%m-%d at %H:%M')
        txt = """\

        {}

        version: {}
        atooms version: {}
        simulation started on: {}
        output path: {}\
        """.format(self, self.version, __version__, now, self.output_path)
        return txt

    def _info_backend(self):
        """Subclasses may want to override this method."""
        txt = 'backend: {}\n'.format(self.backend.__class__)
        if hasattr(self.backend, 'version'):
            txt += 'backend version: {}\n'.format(self.backend.version)
        return txt

    def _info_observers(self):
        txt = []
        for observer in self._observer:
            callback = observer['callback']
            scheduler = observer['scheduler']
            if 'target' in _callable_name(callback):
                args = observer['args']
                txt.append('target %s: %s' % (_callable_name(callback), args[0]))
            else:
                txt.append('writer %s: interval=%s calls=%s' %
                           (_callable_name(callback), scheduler.interval, scheduler.calls))
        return '\n'.join(txt) + '\n'

    def _info_end(self):
        now = datetime.datetime.now().strftime('%Y-%m-%d at %H:%M')
        txt = """
        final steps: {}
        final rmsd: {:.2f}
        wall time [s]: {:.2f}
        average TSP [s/step/particle]: {:.2e}
        simulation ended on: {}\
        """.format(self.current_step, self.rmsd, self.wall_time(),
                   self.wall_time(per_step=True, per_particle=True),
                   now)
        return txt



import os
import re
import glob

from atooms.system.particle import distinct_species
from atooms.trajectory.xyz import TrajectoryXYZ
from atooms.trajectory import SuperTrajectory


class TrajectoryRUMD(TrajectoryXYZ):

    """RUMD trajectory format (https://rumd.org)"""

    # TODO: allow reading unfolded configuration by parsing the box image integers

    def __init__(self, filename, mode='r', fields=None):
        # We do not currently support writing with custom fields
        # The format is: species, position, velocity
        if mode == 'w' and fields is not None:
            raise ValueError('TrajectoryRUMD cannot write custom fields')
        super(TrajectoryRUMD, self).__init__(filename, mode,
                                             alias={'timeStepIndex': 'step',
                                                    'boxLengths': 'cell',
                                                    'sim_box': 'cell'},
                                             fields=fields)
        # The minimum id for RUMD is 0
        self._min_id = 0

    def read_steps(self):
        steps = super(TrajectoryRUMD, self).read_steps()

        # RUMD specific stuff
        basename_ext = os.path.basename(self.filename)
        basename = basename_ext.split('.')[0]

        # For native rumd trajectories we add the block offset
        s = re.search(r'(trajectory|block)(\d+)', basename)
        if s:
            # Redefine samples and steps to make sure these are the absolute step indexes
            # This is important when trajectories are written in blocks.
            _, block = s.group(1), s.group(2)
            iblock = int(block)
            dt = steps[-1]
            steps = [i + dt*iblock for i in steps]

        # If we follow a folder based logic, files are named according
        # to the step and contain a single step like
        #     folder/0000001.xyz.gz
        # We grab the step from
        # the file name.
        s = re.search(r'^(\d+)$', basename)
        if s and len(steps) == 1:
            steps = [int(basename)]
        return steps

    def _read_comment(self, frame):
        meta = super(TrajectoryRUMD, self)._read_comment(frame)

        # RUMD specific stuff that can't be handled as aliases.
        if 'integrator' in meta:
            meta['dt'] = meta['integrator'][1]
        if 'sim_box' in meta:
            # After sim_box there is a keyword for the box type which we ignore
            meta['cell'] = meta['sim_box'][1:]
            meta['ndim'] = len(meta['cell'])

        return meta

    def _comment(self, step, system):

        def first_of_species(system, species):
            for i, p in enumerate(system.particle):
                if p.species == species:
                    return i
            raise ValueError('no species %d found in system' % species)

        sp = distinct_species(system.particle)
        mass = [system.particle[first_of_species(system, isp)].mass for isp in sp]
        hdr = 'ioformat=1 dt=%g timeStepIndex=%d boxLengths=' + \
              '%.12f,%.12f,%.12f' + \
              ' numTypes=%d mass=' + '%g,'*(len(sp))
        hdr = hdr.strip(',')
        hdr += ' columns=type,x,y,z,vx,vy,vz\n'
        return hdr % tuple([self.timestep, step] + list(system.cell.side) + [len(sp)] + mass)

    def write_system(self, system, step):
        sp = distinct_species(system.particle)
        self._file.write("%d\n" % len(system.particle))
        self._file.write(self._comment(step, system))
        ndim = len(system.particle[0].position)
        for p in system.particle:
            # We get the integer index corresponding to species. Ex.:
            # if species are 'A', 'B' we get 0 and 1. Note that in
            # general getting the sample back via read_system() will
            # not preserve the species.
            isp = sp.index(p.species)
            pos = ' '.join([str(x) for x in p.position])
            vel = ' '.join([str(x) for x in p.velocity])
            self._file.write(f"{isp} {pos} {vel}\n")


class SuperTrajectoryRUMD(SuperTrajectory):

    """SuperTrajectory for RUMD format"""

    # TODO: why new here? init should be enough and will make it possible dynamic extension
    def __new__(cls, inp, mode='r', basename='trajectory*.gz'):
        """ Takes a directory as input and get all block*gz files in there """
        if not os.path.isdir(inp):
            raise IOError("We expected this to be a dir (%s)" % inp)
        f_all = glob.glob(inp + '/' + basename)
        if len(f_all) == 0:
            # Let's try with 00000.xyz.gz lie files
            f_all = glob.glob(inp + '/[0-9]*gz')
            f_all.sort()
        elif len(f_all) > 0:
            # Avoid last block because rumd does not write the last cfg!
            f_all.sort()
            f_all = f_all[:-1]
        return SuperTrajectory(f_all, TrajectoryRUMD)



import os
import unittest

try:
    import h5py
    from atooms.trajectory import TrajectoryHDF5
    from atooms.system import System, Cell, Particle, Interaction
    from atooms.trajectory.hdf5 import _CutOff, _PairPotential
    HAS_HDF5 = True
except:
    HAS_HDF5 = False


class Test(unittest.TestCase):

    @unittest.skipIf(not HAS_HDF5, 'no h5py module')
    def test_write_initial_state(self):
        p = [_PairPotential("lennard_jones", {"epsilon": 1.0, "sigma": 1.0},
                            [1, 1], _CutOff("CS", 2.5))]
        i = [Interaction()]
        i[0].name = "atomic"
        i[0].potential = p
        s = System()
        s.particle = [Particle(position=[1.0, 1.0, 1.0],
                               velocity=[0.0, 0.0, 0.0])]
        s.cell = Cell([1.0, 1.0, 1.0])
        with TrajectoryHDF5('/tmp/test_hdf5.h5', 'w') as t:
            t.write_interaction(i)
            t.write(s, 0)

        with TrajectoryHDF5('/tmp/test_hdf5.h5', 'r') as t:
            i = t.read_interaction()
            s = t[0]

    @unittest.skipIf(not HAS_HDF5, 'no h5py module')
    def test_strings(self):
        import numpy
        with h5py.File('/tmp/test_hdf5.h5', 'w') as fh:
            # fh['test'] = ['hello']  # this will fail with python3
            fh['test'] = [b'hello']
        with h5py.File('/tmp/test_hdf5.h5', 'r') as fh:
            self.assertEqual(fh['test'][0].decode(), 'hello')

    def tearDown(self):
        os.remove('/tmp/test_hdf5.h5')


if __name__ == '__main__':
    unittest.main()
